import numpy as np
import logging
from typing import Union, List, Dict, Tuple
from kgcnn.data.transform.scaler.mol import ExtensiveMolecularScalerBase

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class EnergyForceExtensiveLabelScaler(ExtensiveMolecularScalerBase):
    r"""Extensive scaler for scaling jointly energy, forces.

    Inherits from :obj:`kgcnn.scaler.mol.ExtensiveMolecularScalerBase` but makes use of `X` , `y` , as
    `atomic_number` and (`energy` , `force` ).
    In contrast to :obj:`kgcnn.scaler.mol.ExtensiveMolecularLabelScaler` which uses only
    `y` as `energy` .

    Interface is designed after scikit-learn scaler and has additional functions to apply on datasets with
    :obj:`fit_dataset()` and :obj:`transform_dataset()`

    .. note::

        Units for energy and forces must match.

    Code example for scaler:

    .. code-block:: python

        import numpy as np
        from kgcnn.data.transform.scaler.force import EnergyForceExtensiveLabelScaler
        energy = np.random.rand(5).reshape((5,1))
        mol_num = [np.array([6, 1, 1, 1, 1]), np.array([7, 1, 1, 1]),
            np.array([6, 6, 1, 1, 1, 1]), np.array([6, 6, 1, 1]), np.array([6, 6, 1, 1, 1, 1, 1, 1])
        ]
        force = [np.random.rand(len(m)*3).reshape((len(m),3)) for m in mol_num]
        scaler = EnergyForceExtensiveLabelScaler()
        scaler.fit(y=[energy, force], X=mol_num)
        print(scaler.get_weights())
        print(scaler.get_config())
        scaler._plot_predict(energy, mol_num)  # For debugging.
        y, f = scaler.transform(y=[energy, force], X=mol_num)
        print(energy, y)
        print(scaler.inverse_transform(y=[y, f], X=mol_num)[1][1][0], f[0])
        scaler.save("example.json")
        new_scaler = EnergyForceExtensiveLabelScaler()
        new_scaler.load("example.json")
        print(scaler.inverse_transform(y=[y, f], X=mol_num)[1][1][0], f[0])

    """

    def __init__(self, standardize_coordinates: bool = False, **kwargs):
        r"""Initialize layer with arguments for :obj:`kgcnn.scaler.mol.ExtensiveMolecularScalerBase` .

        Args:
            standardize_coordinates (bool): Whether to standardize coordinates. Must always be False.
            kwargs: Kwargs for :obj:`kgcnn.scaler.mol.ExtensiveMolecularScalerBase` parent class.
                See docs for this class.
        """
        super(EnergyForceExtensiveLabelScaler, self).__init__(**kwargs)
        self._standardize_coordinates = standardize_coordinates
        if self._standardize_coordinates:
            raise NotImplementedError("Scaling of coordinates is not supported. This class is a pure label scaler.")
        # Backward compatibility.
        self._use_separate_input_arguments = False

    # noinspection PyPep8Naming
    def fit(self, y: Tuple[List[np.ndarray], List[np.ndarray]] = None, *,
            X: List[np.ndarray] = None,
            sample_weight: Union[None, np.ndarray] = None,
            force: Union[None, List[np.ndarray]] = None,
            atomic_number: Union[None, List[np.ndarray]] = None
            ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Fit Scaler to data.

        Args:
            y (tuple): Tuple of `(energy, forces)` .
                Energies must be a single array or list of energies of shape `(n_samples, n_states)` .
                For one energy this must still be `(n_samples, 1)` . List of forces as with each force stored in a
                numpy array. Note that you can also pass the forces separately to function argument `force` , in
                which case `y` should be only energies (not a tuple).
            X (list): Atomic number `atomic_number` are a list of arrays of atomic numbers.
                Example: `[np.array([7,1,1,1]), ...]` . They must match in length.
                Note that you can also pass the atomic numbers separately to function argument `atomic_number` , in
                which case `X` is ignored.
            sample_weight (list, np.ndarray): Weights for each sample.
            force (list): List of forces as numpy arrays. Deprecated, since they can be contained in `y` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Deprecated, since they can be contained in `X` .

        Returns:
            self.
        """
        X, y, force, atomic_number = self._verify_input(X, y, force, atomic_number)
        return super(EnergyForceExtensiveLabelScaler, self).fit(
            molecular_property=y, sample_weight=sample_weight, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def fit_transform(self, y: Tuple[List[np.ndarray], List[np.ndarray]] = None, *,
                      X: List[np.ndarray] = None,
                      sample_weight: Union[None, np.ndarray] = None,
                      force: Union[None, List[np.ndarray]] = None,
                      atomic_number: Union[None, List[np.ndarray]] = None,
                      copy: bool = True
                      ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Fit Scaler to data and subsequently transform data.

        Args:
            y (tuple): Tuple of `(energy, forces)` .
                Energies must be a single array or list of energies of shape `(n_samples, n_states)` .
                For one energy this must still be `(n_samples, 1)` . List of forces as with each force stored in a
                numpy array. Note that you can also pass the forces separately to function argument `force` , in
                which case `y` should be only energies (not a tuple).
            X (list): Atomic number `atomic_number` are a list of arrays of atomic numbers.
                Example: `[np.array([7,1,1,1]), ...]` . They must match in length.
                Note that you can also pass the atomic numbers separately to function argument `atomic_number` , in
                which case `X` is ignored.
            sample_weight (list, np.ndarray): Weights for each sample.
            force (list): List of forces as numpy arrays. Deprecated, since they can be contained in `y` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Deprecated, since they can be contained in `X` .
            copy (bool): Not yet implemented.

        Returns:
            tuple: Tuple of transformed `(energy, forces)` .
        """
        X, y, force, atomic_number = self._verify_input(X, y, force, atomic_number)
        self.fit(X=X, y=y, atomic_number=atomic_number, force=force, sample_weight=sample_weight)
        return self.transform(X=X, y=y, copy=copy, force=force, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def transform(self, y: Tuple[List[np.ndarray], List[np.ndarray]] = None, *,
                  X: List[np.ndarray] = None,
                  force: Union[None, List[np.ndarray]] = None,
                  atomic_number: Union[None, List[np.ndarray]] = None,
                  copy: bool = True
                  ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Perform scaling of atomic energies and forces.

        Args:
            y (tuple): Tuple of `(energy, forces)` .
                Energies must be a single array or list of energies of shape `(n_samples, n_states)` .
                For one energy this must still be `(n_samples, 1)` . List of forces as with each force stored in a
                numpy array. Note that you can also pass the forces separately to function argument `force` , in
                which case `y` should be only energies (not a tuple).
            X (list): Atomic number `atomic_number` are a list of arrays of atomic numbers.
                Example: `[np.array([7,1,1,1]), ...]` . They must match in length.
                Note that you can also pass the atomic numbers separately to function argument `atomic_number` , in
                which case `X` is ignored.
            force (list): List of forces as numpy arrays. Deprecated, since they can be contained in `y` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Deprecated, since they can be contained in `X` .
            copy (bool): Not yet implemented.

        Returns:
            tuple: Tuple of transformed `(energy, forces)` .
        """
        X, y, force, atomic_number = self._verify_input(X, y, force, atomic_number)
        if copy:
            y = np.array(y) - self.predict(atomic_number)
            if self._standardize_scale:
                y = y / np.expand_dims(self.scale_, axis=0)
                force = [np.array(f) / np.expand_dims(self.scale_, axis=0) for f in force]
            else:
                force = [np.array(f) for f in force]
        else:
            for i in range(len(y)):
                y[i][:] = y[i] - self.predict(atomic_number)[i]
                if self._standardize_scale:
                    y[i][:] = y[i] / self.scale_
                    force[i][:] = force[i] / np.expand_dims(self.scale_, axis=0)
        return y, force

    # noinspection PyPep8Naming
    def inverse_transform(self, y: Tuple[List[np.ndarray], List[np.ndarray]] = None, *,
                          X: List[np.ndarray] = None,
                          force: Union[None, List[np.ndarray]] = None,
                          atomic_number: Union[None, List[np.ndarray]] = None,
                          copy: bool = True
                          ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Scale back data for atoms.

        Args:
            y (tuple): Tuple of `(energy, forces)` .
                Energies must be a single array or list of energies of shape `(n_samples, n_states)` .
                For one energy this must still be `(n_samples, 1)` . List of forces as with each force stored in a
                numpy array. Note that you can also pass the forces separately to function argument `force` , in
                which case `y` should be only energies (not a tuple).
            X (list): Atomic number `atomic_number` are a list of arrays of atomic numbers.
                Example: `[np.array([7,1,1,1]), ...]` . They must match in length.
                Note that you can also pass the atomic numbers separately to function argument `atomic_number` , in
                which case `X` is ignored.
            force (list): List of forces as numpy arrays. Deprecated, since they can be contained in `y` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Deprecated, since they can be contained in `X` .
            copy (bool): Not yet implemented.

        Returns:
            tuple: Tuple of reverse-transformed `(energy, forces)` .
        """
        X, y, force, atomic_number = self._verify_input(X, y, force, atomic_number)
        if copy:
            y = np.array(y)
            if self._standardize_scale:
                y = y * np.expand_dims(self.scale_, axis=0)
                force = [np.array(f) * np.expand_dims(self.scale_, axis=0) for f in force]
            else:
                force = [np.array(f) for f in force]
            y = y + self.predict(atomic_number)
        else:
            for i in range(len(y)):
                if self._standardize_scale:
                    y[i][:] = y[i][:] * self.scale_
                    force[i][:] = force[i] * np.expand_dims(self.scale_, axis=0)
                y[i][:] = y[i][:] + self.predict(atomic_number)[i]
        return y, force

    # Needed for backward compatibility.
    # noinspection PyPep8Naming
    def _verify_input(self, X, y, force, atomic_number):
        # Verify the input format.
        if y is None:
            raise ValueError("`EnergyForceExtensiveLabelScaler` requires 'y' argument, but got 'None'.")
        if force is not None:
            self._use_separate_input_arguments = True
            energy, forces = y, force
        else:
            self._use_separate_input_arguments = False
            energy, forces = y
        if atomic_number is not None:
            atoms = atomic_number
            x_input = X
        else:
            atoms = X
            x_input = None
        return x_input, energy, forces, atoms

    def get_config(self) -> dict:
        """Get configuration for scaler."""
        config = super(EnergyForceExtensiveLabelScaler, self).get_config()
        config.update({"standardize_coordinates": self._standardize_coordinates})
        return config

    def set_config(self, config: dict):
        """Set configuration for scaler.

        Args:
            config (dict): Config dictionary.
        """
        self._standardize_coordinates = config["standardize_coordinates"]
        config_super = {key: value for key, value in config.items() if key not in ["standardize_coordinates"]}
        return super(EnergyForceExtensiveLabelScaler, self).set_config(config_super)

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]], y: List[str] = None, X: str = None,
                    atomic_number: str = None, sample_weight: str = None,  **fit_params):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` containing energies and forces and atomic numbers.
            y (list): List of names for energy and force in dataset. For example ("energy", "forces") .
            X (str): Name of atomic number information in dataset. For example "atomic_number".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic_number".
                Optional, if used, `X` will be ignored.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            fit_params: Fit parameters handed to :obj:`fit()`

        Returns:
            self.
        """
        atoms = atomic_number if atomic_number is not None else X
        energy, force = y
        return self.fit(
            X=[item[atoms] for item in dataset],
            y=([item[energy] for item in dataset], [item[force] for item in dataset]),
            sample_weight=[item[sample_weight] for item in dataset] if sample_weight is not None else None,
            **fit_params
        )

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          X: str = None, y: List[str] = None, atomic_number: str = None, copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` containing energies and forces and atomic numbers.
            y (list): List of names for energy and force in dataset. For example ("energy", "forces") .
            X (str): Name of atomic number information in dataset. For example "atomic_number".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic_number".
                Optional, if used, `X` will be ignored.
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        atoms = atomic_number if atomic_number is not None else X
        energy, force = y
        if copy_dataset:
            dataset = dataset.copy()
        out_energy, out_force = self.transform(
            atomic_number=[graph[atoms] for graph in dataset],
            y=([graph[energy] for graph in dataset], [graph[force] for graph in dataset]),
            copy=copy,
        )
        for graph, graph_energy, graph_force in zip(dataset, out_energy, out_force):
            graph[energy] = graph_energy
            graph[force] = graph_force
        return dataset

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  y: List[str] = None, X: str = None, atomic_number: str = None, copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` containing energies and forces and atomic numbers.
            y (list): List of names for energy and force in dataset. For example ("energy", "forces") .
            X (str): Name of atomic number information in dataset. For example "atomic_number".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic_number".
                Optional, if used, `X` will be ignored.
            copy (bool): Whether to copy dataset. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        atoms = atomic_number if atomic_number is not None else X
        energy, force = y
        if copy_dataset:
            dataset = dataset.copy()
        out_energy, out_force = self.inverse_transform(
            atomic_number=[graph[atoms] for graph in dataset],
            y=([graph[energy] for graph in dataset], [graph[force] for graph in dataset]),
            copy=copy,
        )
        for graph, graph_energy, graph_force in zip(dataset, out_energy, out_force):
            graph[energy] = graph_energy
            graph[force] = graph_force
        return dataset

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              y: List[str] = None, X: str = None,
                              atomic_number: str = None, sample_weight: str = None, copy: bool = True,
                              copy_dataset: bool = False,
                              **fit_params) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` containing energies and forces and atomic numbers.
            y (list): List of names for energy and force in dataset. For example ("energy", "forces") .
            X (str): Name of atomic number information in dataset. For example "atomic_number".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic_number".
                Optional, if used, `X` will be ignored.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy dataset. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.
            fit_params: Fit parameters handed to :obj:`fit()`

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, X=X, y=y, sample_weight=sample_weight, atomic_number=atomic_number,
                         **fit_params)
        return self.transform_dataset(dataset=dataset, X=X, y=y,atomic_number=atomic_number, copy=copy,
                                      copy_dataset=copy_dataset)



import importlib
from typing import Union
from kgcnn.utils.serial import deserialize as deserialize_general

module_list = {
    "StandardScaler": "kgcnn.data.transform.scaler",
    "StandardLabelScaler": "kgcnn.data.transform.scaler",
    "ExtensiveMolecularScaler": "kgcnn.data.transform.mol",
    "ExtensiveMolecularLabelScaler": "kgcnn.data.transform.mol",
    "EnergyForceExtensiveLabelScaler": "kgcnn.data.transform.force",
    "QMGraphLabelScaler": "kgcnn.data.transform.mol"
}


def deserialize(name: Union[str, dict], **kwargs):
    """Deserialize a scaler class.

    Args:
        name (str, dict): Serialization dictionary of class. This can also be a name of former graph functions for
            backward compatibility that now coincides with the processor's default name.
        kwargs: Kwargs for processor initialization, if :obj:`name` is string.

    Returns:
        GraphPreProcessorBase: Instance of graph preprocessor.
    """

    if isinstance(name, dict):
        if "class_name" not in name:
            raise ValueError("Require 'class_name' for deserialization")
        if "module_name" not in name:
            if name["class_name"] in module_list:
                name["module_name"] = module_list[name["class_name"]]
            else:
                raise ValueError("Unknown module name for serialization for '%s'." % name["class_name"])
        if "config" not in name:
            name["config"] = {}
        return deserialize_general(name)

    if isinstance(name, str):
        # if given as string name. Lookup identifier.
        if name not in module_list:
            raise ValueError("Unknown name for scaler '%s'." % name)
        module_name = module_list[name]
        obj_class = getattr(importlib.import_module(str(module_name)), str(name))
        return obj_class(**kwargs)

    raise TypeError("Wrong type for deserialization. Require 'str' or 'dict'.")


"""
Module for handling Visual Graph Datasets (VGD).
"""
import os
import typing as t

import numpy as np

from visual_graph_datasets.config import Config
from visual_graph_datasets.util import get_dataset_path, ensure_folder
from visual_graph_datasets.web import get_file_share
from visual_graph_datasets.web import PROVIDER_CLASS_MAP, AbstractFileShare
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf

from kgcnn.data.base import MemoryGraphDataset
from kgcnn.graph.base import GraphDict


class VisualGraphDataset(MemoryGraphDataset):
    r"""Visual graph datasets from external :obj:`visual_graph_datasets` python package.

    Downloading and processing is given in :obj:`visual_graph_datasets` python package.
    For documentation please find `<https://github.com/awa59kst120df/visual_graph_datasets>`_ .
    This class constitutes an interface to kgcnn :obj:`MemoryGraphDataset` data representation.
    """
    
    def __init__(self,
                 name: str,
                 config: Config = Config()):
        super(VisualGraphDataset, self).__init__(
            dataset_name=name,
            file_directory=None,
            file_name=None,
            data_directory=None
        )
        self.vgd_config = config
        self.index_data_map: t.Dict[int, dict] = {}

        self._has_importances: t.Dict[str, bool] = {}

    def ensure(self) -> None:
        """
        This method ensures that the raw dataset exists on the disk. After this method has been called it
        can be certain that the dataset folder exists on the disk and that the folder path is known.

        This is either because an existing dataset folder has been found or because the dataset was
        downloaded.

        Returns:
            None
        """
        # First of all we try to load the dataset, as it might already exist on the system.
        if os.path.exists(os.path.join(self.vgd_config.get_datasets_path(), self.dataset_name)):
            # This function will try to locate a dataset with the given name inside the system's global
            # default folder where all the visual graph datasets are stored. If it does not find a
            # corresponding dataset folder there, an exception is raised.
            self.data_directory = get_dataset_path(
                dataset_name=self.dataset_name,
                datasets_path=self.vgd_config.get_datasets_path()
            )
            return

        # At this point we know that the folder does not already exist which means we need to download the
        # dataset.

        # For this we will first check if a dataset with the given name is even available at the remote
        # file share provider.
        ensure_folder(self.vgd_config.get_datasets_path())
        file_share: AbstractFileShare = get_file_share(self.vgd_config)
        file_share.check_dataset(self.dataset_name)

        # If the dataset is available, then we can download it and then finally load the path
        file_share.download_dataset(self.dataset_name, self.vgd_config.get_datasets_path())
        self.data_directory = get_dataset_path(
            dataset_name=self.dataset_name,
            datasets_path=self.vgd_config.get_datasets_path(),
        )
        self.logger.info(f'visual graph dataset found @ {self.data_directory}')

    def read_in_memory(self) -> None:
        """
        Actually loads the dataset from the file representations into the working memory as GraphDicts
        within the internal MemoryGraphList.

        Returns:
            None
        """
        name_data_map, self.index_data_map = load_visual_graph_dataset(
            self.data_directory,
            logger=self.logger,
            metadata_contains_index=True
        )
        dataset_length = len(self.index_data_map)

        self.empty(dataset_length)
        self.logger.info(f'initialized empty list with {len(self.index_data_map)} elements')

        for index, data in sorted(self.index_data_map.items(), key=lambda t: t[0]):
            g = data['metadata']['graph']

            # In the visual_graph_dataset framework, the train and test split indications are part of the
            # metadata and not the graph itself. We need to set them as graph properties with these
            # specific names however, such that later on the existing base method "get_train_test_indices"
            # of the dataset class can be used.
            if 'train_split' in data['metadata']:
                g['train'] = data['metadata']['train_split']
            if 'test_split' in data['metadata']:
                g['test'] = data['metadata']['test_split']

            # Otherwise the basic structure of the dict from the visual graph dataset should be compatible
            # such that it can be directly used as a GraphDict
            graph_dict = GraphDict(g)
            self[index] = graph_dict

        self.logger.info(f'loaded dataset as MemoryGraphList')

    def visualize_importances(self,
                              output_path: str,
                              gt_importances_suffix: str,
                              node_importances_list: t.List[np.ndarray],
                              edge_importances_list: t.List[np.ndarray],
                              indices: t.List[int],
                              title: str = 'Model',
                              ) -> None:
        data_list = [self.index_data_map[index] for index in indices]

        suffix = gt_importances_suffix
        create_importances_pdf(
            output_path=output_path,
            graph_list=[data['metadata']['graph'] for data in data_list],
            image_path_list=[data['image_path'] for data in data_list],
            node_positions_list=[data['metadata']['graph']['image_node_positions'] for data in data_list],
            importances_map={
                'Ground Truth': (
                    [data['metadata']['graph'][f'node_importances_{suffix}'] for data in data_list],
                    [data['metadata']['graph'][f'edge_importances_{suffix}'] for data in data_list],
                ),
                title: (
                    node_importances_list,
                    edge_importances_list
                )
            }
        )

    # This method loops through all the elements of the dataset which makes it quite computationally
    # expensive, which is why the value will be cached to be more efficient if it is called multiple
    # time redundantly
    def has_importances(self, suffix: t.Union[str, int]) -> bool:
        """
        Returns whether the dataset contains node and edge importances (explanation) annotations which fit
        the given ``suffix``.

        A visual graph dataset may have multiple different importance annotations. The main difference these
        may have is the number of individual importance channels in which the explanations are represented.
        The number of channels is added as a suffix. For example a regression dataset may have the suffix
        "1" which are very simple one-dimensional explanations or the suffix "2" where explanations are
        split into 2 channels where one channel explains the positively influencing input elements and the
        other channel the negatively influencing elements.

        Args:
            suffix: The string suffix of importance annotations to check for.

        Returns:
            boolean
        """
        suffix = str(suffix)

        if suffix not in self._has_importances.keys():
            node_condition = all([f'node_importances_{suffix}' in data['metadata']['graph']
                                  for data in self.index_data_map.values()])

            edge_condition = all([f'edge_importances_{suffix}' in data['metadata']['graph']
                                  for data in self.index_data_map.values()])

            self._has_importances[suffix] = (node_condition and edge_condition)

        return self._has_importances[suffix]

    def get_importances(self,
                        suffix: str,
                        indices: t.Optional[t.List[int]] = None
                        ) -> t.Tuple[t.List[np.ndarray], t.List[np.ndarray]]:
        if indices is None:
            indices = list(self.index_data_map.keys())

        node_importances_list: t.List[np.ndarray] = []
        edge_importances_list: t.List[np.ndarray] = []
        for index in indices:
            data = self.index_data_map[index]
            g = data['metadata']['graph']
            node_importances_list.append(g[f'node_importances_{suffix}'])
            edge_importances_list.append(g[f'edge_importances_{suffix}'])

        return node_importances_list, edge_importances_list

    def __hash__(self):
        return hash(self.dataset_name)



import os
import numpy as np
import scipy.io
import pandas as pd

from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset
from kgcnn.mol.io import write_list_to_xyz_file
from kgcnn.graph.geom import coulomb_matrix_to_inverse_distance_proton, coordinates_from_distance_matrix
from kgcnn.graph.adj import invert_distance
from kgcnn.mol.methods import inverse_global_proton_dict


class QM7bDataset(QMDataset, DownloadDataset):
    r"""Store and process QM7b dataset from `Quantum Machine <http://quantum-machine.org/datasets/>`__ .

    From `Quantum Machine <http://quantum-machine.org/datasets/>`__ :
    This dataset is an extension of the QM7 dataset for multitask learning where 13 additional properties
    (e.g. polarizability, HOMO and LUMO eigenvalues, excitation energies) have to be predicted at different
    levels of theory (ZINDO, SCS, PBE0, GW). Additional molecules comprising chlorine atoms are also included,
    totalling 7211 molecules.

    The dataset is composed of two multidimensional arrays X (7211 x 23 x 23) and T (7211 x 14) representing the inputs
    (Coulomb matrices) and the labels (molecular properties) and one array names of size 14 listing the names of the
    different properties.

    Here, the Coulomb matrices are converted back into coordinates and with :obj:`QMDataset` to molecular structure.
    Labels are not scaled but have original units.

    References:

        (1) L. C. Blum, J.-L. Reymond, 970 Million Druglike Small Molecules for Virtual Screening in
            the Chemical Universe Database GDB-13, J. Am. Chem. Soc., 131:8732, 2009.
        (2) G. Montavon, M. Rupp, V. Gobre, A. Vazquez-Mayagoitia, K. Hansen, A. Tkatchenko, K.-R. Müller,
            O.A. von Lilienfeld, Machine Learning of Molecular Electronic Properties in Chemical Compound Space,
            New J. Phys. 15 095003, 2013.

    """

    download_info = {
        "dataset_name": "QM7b",
        "data_directory_name": "qm7b",
        # https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7b.mat
        "download_url": "http://quantum-machine.org/data/qm7b.mat",
        "download_file_name": 'qm7b.mat',
        "unpack_tar": False,
        "unpack_zip": False,
    }

    def __init__(self, reload: bool = False, verbose: int = 10):
        """Initialize QM9 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM7b")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = ["aepbe0", "zindo-excitation-energy-with-the-most-absorption",
                            "zindo-highest-absorption", "zindo-homo", "zindo-lumo",
                            "zindo-1st-excitation-energy", "zindo-ionization-potential", "zindo-electron-affinity",
                            "ks-homo", "ks-lumo", "gw-homo", "gw-lumo", "polarizability-pbe", "polarizability-scs"]
        self.label_units = ["[?]"] * 14
        self.label_unit_conversion = np.array([1.0] * 14)  # Pick always same units for training
        self.dataset_name = "QM7b"
        self.require_prepare_data = True
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_name = "qm7b.csv"

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)

    def prepare_data(self, overwrite: bool = False, file_column_name: str = None, make_sdf: bool = True):

        if not os.path.exists(self.file_path_xyz) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            coulomb_mat = mat["X"]
            graph_len = [int(np.around(np.sum(np.diagonal(x) > 0))) for x in coulomb_mat]
            proton_inv_dist = [coulomb_matrix_to_inverse_distance_proton(x[:i, :i], unit_conversion=0.529177210903) for
                               x, i in zip(coulomb_mat, graph_len)]
            proton = [x[1] for x in proton_inv_dist]
            inv_dist = [x[0] for x in proton_inv_dist]
            dist = [invert_distance(x) for x in inv_dist]
            pos = [coordinates_from_distance_matrix(x) for x in dist]
            atoms = [[inverse_global_proton_dict[i] for i in x] for x in proton]
            atoms_pos = [[x, y] for x, y in zip(atoms, pos)]
            self.info("Writing XYZ file from coulomb matrix information.")
            write_list_to_xyz_file(self.file_path_xyz, atoms_pos)
        else:
            self.info("Found XYZ file for qm7b already created.")

        if not os.path.exists(self.file_path) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            labels = mat["T"]
            targets = pd.DataFrame(labels, columns=self.label_names)
            self.info("Writing CSV file of graph labels.")
            targets.to_csv(self.file_path, index=False)
        else:
            self.info("Found CSV file of graph labels.")

        return super(QM7bDataset, self).prepare_data(
            overwrite=overwrite, file_column_name=file_column_name, make_sdf=make_sdf)



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectDielectricDataset(MatBenchDataset2020):
    """Store and process :obj:`MatProjectDielectricDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database. Name within Matbench: 'matbench_dielectric'.

    Matbench test dataset for predicting refractive index from structure. Adapted from Materials Project database.
    Removed entries having a formation energy (or energy above the convex hull) more than 150meV and
    those having refractive indices less than 1 and those containing noble gases. Retrieved April 2, 2019.
    For benchmarking w/ nested cross validation, the order of the dataset must be identical to the retrieved data;
    refer to the Automatminer/Matbench publication for more details.

        * Number of samples: 4764
        * Task type: regression
        * Input type: structure

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectDielectricDataset, self).__init__("matbench_dielectric", reload=reload, verbose=verbose)
        self.label_names = "n_r"
        self.label_units = ""



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    name="kgcnn",
    version="2.2.4",  # If version is updated, change version in `kgcnn.__init__` too. (and update changelog)
    author="Patrick Reiser",
    author_email="patrick.reiser@kit.edu",
    description="General Base Layers for Graph Convolutions with tensorflow.keras",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aimat-lab/gcnn_keras",
    install_requires=[
        "numpy>=1.23.0",
        "scikit-learn>=1.1.3",
        "pandas>=1.5.2",
        "scipy>=1.9.3",
        "matplotlib>=3.6.0",
        "rdkit>=2022.9.2",
        "pymatgen>=2022.11.7",
        "tensorflow>=2.9.0",
        # "tensorflow-gpu>=2.10.0",  # No longer used in tensorflow.
        "tensorflow-addons>=0.18.0",
        "keras-tuner>=1.1.3",
        "requests>=2.28.1",
        "networkx>=2.8.8",
        "sympy>=1.11.1",
        "pyyaml>=6.0",
        "ase>=3.22.1",
        "click>=7.1.2",
        "visual_graph_datasets>=0.7.1",
        "brotli>=1.0.9"
    ],
    extras_require={
        "openbabel": ["openbabel"],
    },
    packages=find_packages(),
    include_package_data=True,
    package_data={"kgcnn": ["*.json", "*.yaml"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    keywords=["materials", "science", "machine", "learning", "deep", "graph", "networks", "neural"]
)



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectIsMetalDataset(MatBenchDataset2020):
    r"""Store and process :obj:`MatProjectIsMetalDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database. Name within Matbench: 'matbench_mp_is_metal'.

    Matbench test dataset for predicting DFT metallicity from structure. Adapted from Materials Project database.
    Removed entries having a formation energy (or energy above the convex hull) more than 150meV and those containing
    noble gases. Retrieved April 2, 2019. For benchmarking w/ nested cross validation, the order of the dataset must
    be identical to the retrieved data; refer to the Automatminer/Matbench publication for more details.

        * Number of samples: 106113.
        * Task type: classification.
        * Input type: structure.

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectIsMetalDataset, self).__init__("matbench_mp_is_metal", reload=reload, verbose=verbose)
        self.label_names = "is_metal"
        self.label_units = ""



import typing as t

import numpy as np
import tensorflow as tf

from tensorflow.python.keras.engine import compile_utils
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.modules import Dense, OptionalInputEmbedding
from kgcnn.layers.modules import Activation, Dropout
from kgcnn.layers.modules import LazyConcatenate, LazyAverage
from kgcnn.layers.conv.gat_conv import MultiHeadGATV2Layer
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.pooling import PoolingWeightedNodes, PoolingNodes
from kgcnn.xai.base import ImportanceExplanationMixin

ks = tf.keras

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2022.12.15"


def shifted_sigmoid(x, shift=5, multiplier=1):
    return ks.backend.sigmoid((x - shift) / multiplier)


class ExplanationSparsityRegularization(GraphBaseLayer):

    def __init__(self,
                 factor: float = 1.0,
                 **kwargs):
        super(ExplanationSparsityRegularization, self).__init__(**kwargs)
        self.factor = factor

    def call(self, inputs, **kwargs):
        r"""Computes a loss from importance scores.

        Args:
            inputs: Importance tensor of shape ([batch], [N], K) .

        Returns:
            None.
        """
        # importances: ([batch], [N], K)
        importances = inputs

        loss = tf.reduce_mean(tf.math.abs(importances))
        self.add_loss(loss * self.factor)


class MEGAN(ks.models.Model, ImportanceExplanationMixin):
    """
    MEGAN: Multi Explanation Graph Attention Network

    This model currently supports graph regression and graph classification problems. It was mainly designed
    with a focus on explainable AI (XAI). Along the main prediction, this model is able to output multiple
    attention-based explanations for that prediction. More specifically, the model outputs node and edge
    attributional explanations (assigning [0, 1] values to ever node / edge of the input graph) in K
    separate explanation channels, where K can be chosen as an independent model parameter.
    """
    __kgcnn_model_version__ = __model_version__

    def __init__(self,
                 # convolutional network related arguments
                 units: t.List[int],
                 activation: str = "kgcnn>leaky_relu",
                 use_bias: bool = True,
                 dropout_rate: float = 0.0,
                 use_edge_features: bool = True,
                 input_embedding: dict = None,
                 # node/edge importance related arguments
                 importance_units: t.List[int] = [],
                 importance_channels: int = 2,
                 importance_activation: str = "sigmoid",  # do not change
                 importance_dropout_rate: float = 0.0,  # do not change
                 importance_factor: float = 0.0,
                 importance_multiplier: float = 10.0,
                 sparsity_factor: float = 0.0,
                 concat_heads: bool = True,
                 # mlp tail end related arguments
                 final_units: t.List[int] = [1],
                 final_dropout_rate: float = 0.0,
                 final_activation: str = 'linear',
                 final_pooling: str = 'sum',
                 regression_limits: t.Optional[t.Tuple[float, float]] = None,
                 regression_reference: t.Optional[float] = None,
                 return_importances: bool = True,
                 **kwargs):
        """
        Args:
            units: A list of ints where each element configures an additional attention layer. The numeric
                value determines the number of hidden units to be used in the attention heads of that layer
            activation: The activation function to be used within the attention layers of the network
            use_bias: Whether the layers of the network should use bias weights at all
            dropout_rate: The dropout rate to be applied after *each* of the attention layers of the network.
            input_embedding: Dictionary of embedding kwargs for input embedding layer.
            use_edge_features: Whether edge features should be used. Generally the network supports the
                usage of edge features, but if the input data does not contain edge features, this should be
                set to False.
            importance_units: A list of ints where each element configures another dense layer in the
                subnetwork that produces the node importance tensor from the main node embeddings. The
                numeric value determines the number of hidden units in that layer.
            importance_channels: The int number of explanation channels to be produced by the network. This
                is the value referred to as "K". Note that this will also determine the number of attention
                heads used within the attention subnetwork.
            importance_factor: The weight of the explanation-only train step. If this is set to exactly
                zero then the explanation train step will not be executed at all (less computationally
                expensive)
            importance_multiplier: An additional hyperparameter of the explanation-only train step. This
                is essentially the scaling factor that is applied to the values of the dataset such that
                the target values can reasonably be approximated by a sum of [0, 1] importance values.
            sparsity_factor: The coefficient for the sparsity regularization of the node importance
                tensor.
            concat_heads: Whether to concat the heads of the attention subnetwork. The default is True. In
                that case the output of each individual attention head is concatenated and the concatenated
                vector is then used as the input of the next attention layer's heads. If this is False, the
                vectors are average pooled instead.
            final_units: A list of ints where each element configures another dense layer in the MLP
                at the tail end of the network. The numeric value determines the number of the hidden units
                in that layer. Note that the final element in this list has to be the same as the dimension
                to be expected for the samples of the training dataset!
            final_dropout_rate: The dropout rate to be applied after *every* layer of the final MLP.
            final_activation: The activation to be applied at the very last layer of the MLP to produce the
                actual output of the network.
            final_pooling: The pooling method to be used during the global pooling phase in the network.
            regression_limits: A tuple where the first value is the lower limit for the expected value range
                of the regression task and teh second value the upper limit.
            regression_reference: A reference value which is inside the range of expected values (best if
                it was in the middle, but does not have to). Choosing different references will result
                in different explanations.
            return_importances: Whether the importance / explanation tensors should be returned as an output
                of the model. If this is True, the output of the model will be a 3-tuple:
                (output, node importances, edge importances), otherwise it is just the output itself
        """
        super(MEGAN, self).__init__(self, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.dropout_rate = dropout_rate
        self.use_edge_features = use_edge_features
        self.input_embedding = input_embedding
        self.importance_units = importance_units
        self.importance_channels = importance_channels
        self.importance_activation = importance_activation
        self.importance_dropout_rate = importance_dropout_rate
        self.importance_factor = importance_factor
        self.importance_multiplier = importance_multiplier
        self.sparsity_factor = sparsity_factor
        self.concat_heads = concat_heads
        self.final_units = final_units
        self.final_dropout_rate = final_dropout_rate
        self.final_activation = final_activation
        self.final_pooling = final_pooling
        self.regression_limits = regression_limits
        self.regression_reference = regression_reference
        self.return_importances = return_importances

        # ~ MAIN CONVOLUTIONAL / ATTENTION LAYERS
        if self.input_embedding:
            self.embedding_nodes = OptionalInputEmbedding(**self.input_embedding["node"])
        self.attention_layers: t.List[GraphBaseLayer] = []
        for u in self.units:
            lay = MultiHeadGATV2Layer(
                units=u,
                num_heads=self.importance_channels,
                use_edge_features=self.use_edge_features,
                activation=self.activation,
                use_bias=self.use_bias,
                has_self_loops=True,
                concat_heads=self.concat_heads
            )
            self.attention_layers.append(lay)

        self.lay_dropout = Dropout(rate=self.dropout_rate)

        # ~ EDGE IMPORTANCES
        self.lay_act_importance = Activation(activation=self.importance_activation)
        self.lay_concat_alphas = LazyConcatenate(axis=-1)

        self.lay_pool_edges_in = PoolingLocalEdges(pooling_method='mean', pooling_index=0)
        self.lay_pool_edges_out = PoolingLocalEdges(pooling_method='mean', pooling_index=1)
        self.lay_average = LazyAverage()

        # ~ NODE IMPORTANCES
        self.node_importance_units = importance_units + [self.importance_channels]
        self.node_importance_acts = ['relu' for _ in importance_units] + ['linear']
        self.node_importance_layers = []
        for u, act in zip(self.node_importance_units, self.node_importance_acts):
            lay = Dense(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.node_importance_layers.append(lay)
            
        self.lay_sparsity = ExplanationSparsityRegularization(factor=self.sparsity_factor)

        # ~ OUTPUT / MLP TAIL END
        self.lay_pool_out = PoolingNodes(pooling_method=self.final_pooling)
        self.lay_concat_out = LazyConcatenate(axis=-1)
        self.lay_final_dropout = Dropout(rate=self.final_dropout_rate)

        self.final_acts = ["relu" for _ in self.final_units]
        self.final_acts[-1] = self.final_activation
        self.final_biases = [True for _ in self.final_units]
        self.final_biases[-1] = False
        self.final_layers = []
        for u, act, bias in zip(self.final_units, self.final_acts, self.final_biases):
            lay = Dense(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.final_layers.append(lay)

        # ~ EXPLANATION ONLY TRAIN STEP
        self.bce_loss = ks.losses.BinaryCrossentropy()
        self.compiled_classification_loss = compile_utils.LossesContainer(self.bce_loss)

        self.mse_loss = ks.losses.MeanSquaredError()
        self.mae_loss = ks.losses.MeanAbsoluteError()
        self.compiled_regression_loss = compile_utils.LossesContainer(self.mae_loss)

        if self.regression_limits is not None:
            self.regression_width = np.abs(self.regression_limits[1] - self.regression_limits[0])

    def get_config(self):
        config = super(MEGAN, self).get_config()
        config.update({
            "units": self.units,
            "activation": self.activation,
            "use_bias": self.use_bias,
            "dropout_rate": self.dropout_rate,
            "use_edge_features": self.use_edge_features,
            "importance_units": self.importance_units,
            "importance_channels": self.importance_channels,
            "importance_activation": self.importance_activation,
            "importance_dropout_rate": self.importance_dropout_rate,
            "importance_factor": self.importance_factor,
            "importance_multiplier": self.importance_multiplier,
            "sparsity_factor": self.sparsity_factor,
            "concat_heads": self.concat_heads,
            "final_units": self.final_units,
            "final_dropout_rate": self.final_dropout_rate,
            "final_activation": self.final_activation,
            "final_pooling": self.final_pooling,
            "regression_limits": self.regression_limits,
            "regression_reference": self.regression_reference,
            "return_importances": self.return_importances,
            "input_embedding": self.input_embedding
        })

        return config

    @property
    def doing_regression(self) -> bool:
        return self.regression_limits is not None

    def call(self,
             inputs,
             training: bool = False,
             return_importances: bool = False):

        node_input, edge_input, edge_index_input = inputs

        if self.input_embedding:
            node_input = self.embedding_nodes(node_input, training=training)
        # First of all we apply all the graph convolutional / attention layers. Each of those layers outputs
        # the attention logits alpha additional to the node embeddings. We collect all the attention logits
        # in a list so that we can later sum them all up.
        alphas = []
        x = node_input
        for lay in self.attention_layers:
            # x: ([batch], [N], F)
            # alpha: ([batch], [M], K, 1)
            x, alpha = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x, training=training)

            alphas.append(alpha)

        # We sum up all the individual layers attention logit tensors and the edge importances are directly
        # calculated by applying a sigmoid on that sum.
        alphas = self.lay_concat_alphas(alphas)
        edge_importances = tf.reduce_sum(alphas, axis=-1, keepdims=False)
        edge_importances = self.lay_act_importance(edge_importances)

        # Part of the final node importance tensor is actually the pooled edge importances, so that is what
        # we are doing here. The caveat here is that we assume undirected edges as two directed edges in
        # opposing direction. To now achieve a symmetric pooling of these edges we have to pool in both
        # directions and then use the average of both.
        pooled_edges_in = self.lay_pool_edges_in([node_input, edge_importances, edge_index_input])
        pooled_edges_out = self.lay_pool_edges_out([node_input, edge_importances, edge_index_input])
        pooled_edges = self.lay_average([pooled_edges_out, pooled_edges_in])

        node_importances_tilde = x
        for lay in self.node_importance_layers:
            node_importances_tilde = lay(node_importances_tilde)

        node_importances_tilde = self.lay_act_importance(node_importances_tilde)

        node_importances = node_importances_tilde * pooled_edges
        self.lay_sparsity(node_importances)

        # Here we apply the global pooling. It is important to note that we do K separate pooling operations
        # were each time we use the same node embeddings x but a different slice of the node importances as
        # the weights! We concatenate all the individual results in the end.
        outs = []
        for k in range(self.importance_channels):
            node_importance_slice = tf.expand_dims(node_importances[:, :, k], axis=-1)
            out = self.lay_pool_out(x * node_importance_slice)

            outs.append(out)

        # out: ([batch], F*K)
        out = self.lay_concat_out(outs)

        # Now "out" is a graph embedding vector of known dimension so we can simply apply the normal dense
        # mlp to get the final output value.
        for lay in self.final_layers:
            out = lay(out)
            if training:
                self.lay_final_dropout(out, training=training)

        if self.doing_regression:
            reference = tf.ones_like(out) * self.regression_reference
            out = out + reference

        # Usually, the node and edge importance tensors would be direct outputs of the model as well, but
        # we need the option to just return the output alone to be compatible with the standard model
        # evaluation pipeline already implemented in the library.
        if self.return_importances or return_importances:
            return out, node_importances, edge_importances
        else:
            return out

    def regression_augmentation(self,
                                out_true: tf.RaggedTensor):
        """
        Given the tensor ([B], 1) of true regression target values, this method will return two derived
        tensors: The first one is a ([B], 2) tensor of normalized distances of the corresponding true
        values to ``self.regression_reference`` and the second is a ([B], 2) boolean mask tensor.
        Args:
            out_true: A tensor of shape ([B], 1) of the true target values of the current batch.
        Returns:
            A tuple of two tensors each with the shape ([B], 2)
        """
        center_distances = tf.abs(out_true - self.regression_reference)
        center_distances = (center_distances * self.importance_multiplier) / (
                0.5 * self.regression_width)

        # So we need two things: a "samples" tensor and a "mask" tensor. We are going to use the samples
        # tensor as the actual ground truth which acts as the regression target during the explanation
        # train step. The binary values of the mask will determine at which positions a loss should
        # actually be calculated for both of the channels

        # The "lower" part is all the samples which have a target value below the reference value.
        lo_mask = tf.where(out_true < self.regression_reference, 1.0, 0.0)
        # The "higher" part all of the samples above reference
        hi_mask = tf.where(out_true > self.regression_reference, 1.0, 0.0)

        samples = tf.concat([center_distances, center_distances], axis=-1)
        mask = tf.concat([lo_mask, hi_mask], axis=-1)

        return samples, mask

    def train_step(self, data):
        """The method which is called for each individual batch of the training step. In this method a forward
        pass of the input data of the batch is performed, the loss is calculated, the gradients of this loss
        w.r.t to the model parameters are calculated and these gradients are used to update the weights of
        the model.

        This custom train step implements the "explanation co-training": Additionally to the main prediction
        loss there is also a loss term which uses only the node importance values to approximate a
        normalized versions of the target values.

        Args:
            data: Tuple of two tensors (X, Y) where X is the batch tensor of all the inputs and Y is the
                batch tensor of the target output values.

        Returns:
            A dictionary whose keys are the names of metrics and the values are the corresponding values
            of these metrics for this iteration of the training.

        """
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        with tf.GradientTape() as tape:
            if self.return_importances:
                out_true, ni_true, ei_true = y
            else:
                out_true = y

            y_pred = self(x, training=True, return_importances=True)
            out_pred, ni_pred, ei_pred = y_pred
            loss = self.compiled_loss(
                y,
                y_pred if self.return_importances else out_pred,
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )

            exp_metrics = {}
            if self.importance_factor != 0:
                # ~ explanation loss
                # First of all we need to assemble the approximated model output, which is simply calculated
                # by applying a global pooling operation on the corresponding slice of the node importances.
                # So for each slice (each importance channel) we get a single value, which we then
                # concatenate into an output vector with K dimensions.
                outs = []
                for k in range(self.importance_channels):
                    node_importances_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)
                    out = self.lay_pool_out(node_importances_slice)

                    outs.append(out)

                # outs: ([batch], K)
                outs = self.lay_concat_out(outs)

                if self.doing_regression:
                    out_true, mask = self.regression_augmentation(out_true)
                    out_pred = outs
                    exp_loss = self.importance_channels * self.compiled_regression_loss(out_true * mask,
                                                                                        out_pred * mask)

                else:
                    out_pred = shifted_sigmoid(
                        outs,

                        # shift=self.importance_multiplier,
                        # multiplier=(self.importance_multiplier / 5)

                        shift=self.importance_multiplier,
                        multiplier=1,
                    )
                    exp_loss = self.compiled_classification_loss(out_true, out_pred)

                exp_loss *= self.importance_factor
                exp_metrics['exp_loss'] = exp_loss
                loss += exp_loss

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        self.compiled_metrics.update_state(
            y,
            out_pred if not self.return_importances else [out_pred, ni_pred, ei_pred],
            sample_weight=sample_weight
        )

        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {
            **{m.name: m.result() for m in self.metrics},
            **exp_metrics
        }

    def explain_importances(self,
                            x: t.Sequence[tf.Tensor],
                            **kwargs
                            ) -> t.Tuple[tf.RaggedTensor, tf.RaggedTensor]:
        y, node_importances, edge_importances = self(x, return_importances=True)
        return node_importances, edge_importances


def make_model(inputs: t.Optional[list] = None,
               **kwargs
               ):
    r"""Functional model definition of MEGAN. Please check documentation of :obj:`kgcnn.literature.MEGAN` .

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        kwargs: Kwargs for MEGAN model. Please check documentation of :obj:`kgcnn.literature.MEGAN` .

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Building the actual model
    megan = MEGAN(**kwargs)

    # Wrapping the actual model inside a keras functional model to be able to account for the input shapes
    # definitions which are provided.
    layer_inputs = [ks.layers.Input(**kwargs) for kwargs in inputs]
    outputs = megan(layer_inputs)
    model = ks.models.Model(inputs=layer_inputs, outputs=outputs)

    model.__kgcnn_model_version__ = __model_version__
    return model



import os
import numpy as np
import pandas as pd
from kgcnn.data.base import MemoryGraphDataset
from kgcnn.data.download import DownloadDataset


class MD17RevisedDataset(DownloadDataset, MemoryGraphDataset):
    r"""Store and process trajectories from :obj:`MD17DatasetRevised` dataset.

    The information of the readme file is given below:

    The molecules are taken from the original MD17 dataset by
    `Chmiela et al. <https://www.science.org/doi/10.1126/sciadv.1603015>`__ , and 100,000 structures are taken,
    and the energies and forces are recalculated at the PBE/def2-SVP level of theory using very tight SCF convergence
    and very dense DFT integration grid. As such, the dataset is practically free from nummerical noise.

    One warning: As the structures are taken from a molecular dynamics simulation
    (i.e. time series data), they are not guaranteed to be independent samples.
    This is easily evident from the auto-correlation function for the original MD17 dataset

    In short: DO NOT train a model on more than 1000 samples from this dataset.
    Data already published with 50K samples on the original MD17 dataset should be considered
    meaningless due to this fact and due to the noise in the original data.

    The data:
    The ten molecules are saved in Numpy .npz format.
    The keys correspond to:

        - 'nuclear_charges'   : The nuclear charges for the molecule
        - 'coords'            : The coordinates for each conformation (in units of Angstrom)
        - 'energies'          : The total energy of each conformation (in units of kcal/mol)
        - 'forces'            : The cartesian forces of each conformation (in units of kcal/mol/Angstrom)
        - 'old_indices'       : The index of each conformation in the original MD17 dataset
        - 'old_energies'      : The energy of each conformation taken from the original MD17 dataset
        - 'old_forces'        : The forces of each conformation taken from the original MD17 dataset

    Note that for Azobenzene, only 99988 samples are available due to 11 failed DFT calculations due to van der
    Walls clash, and the original dataset only contained 99999 structures.

    Data splits:
    Five training and test splits are saved in CSV format containing the corresponding indices.

    References:

        (1) Anders Christensen, O. Anatole von Lilienfeld, Revised MD17 dataset, Materials Cloud Archive 2020.82 (2020),
            doi: 10.24435/materialscloud:wy-kn.

    """

    download_info = {
        "dataset_name": "MD17Revised",
        "data_directory_name": "MD17Revised",
        "download_url": "https://archive.materialscloud.org/record/file?filename=rmd17.tar.bz2&record_id=466",
        "download_file_name": 'rmd17.tar.bz2',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "rmd17"
    }
    possible_trajectory_names = ["aspirin", "azobenzene", "benzene", "ethanol", "malonaldehyde", "naphthalene",
                                 "paracetamol", "salicylic", "toluene", "uracil"]

    def __init__(self, trajectory_name: str = None, reload=False, verbose=10):
        """Initialize MD17DatasetRevised dataset for a specific trajectory.

        Args:
            trajectory_name (str): Name of trajectory to load.
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        self.data_keys = None
        self.trajectory_name = trajectory_name
        if trajectory_name not in self.possible_trajectory_names:
            raise ValueError(
                "Name for trajectory '%s' not found. Choose: %s." % (trajectory_name, self.possible_trajectory_names))

        MemoryGraphDataset.__init__(self, dataset_name="MD17Revised", verbose=verbose)
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name,
                                           self.unpack_directory_name, "rmd17")
        self.file_name = "rmd17_%s.npz" % self.trajectory_name
        self.dataset_name = self.dataset_name + "_" + self.trajectory_name
        # May add name of trajectory to name of dataset here.

        if self.fits_in_memory:
            self.read_in_memory()

    def _get_trajectory_from_npz(self, file_path: str = None):
        if file_path is None:
            file_dir = os.path.join(self.data_directory, "npz_data")
            file_path = os.path.join(file_dir, self.file_name)
        return np.load(file_path)

    def _get_train_test_splits(self):
        file_dir = os.path.join(self.data_directory, "splits")

        def read_splits(file_name: str) -> list:
            return [
                np.squeeze(pd.read_csv(
                    os.path.join(file_dir, file_name % i), header=None).values, axis=-1) for i in range(1, 6)]
        return read_splits("index_train_0%i.csv"), read_splits("index_test_0%i.csv")

    def read_in_memory(self,):
        """Read dataset trajectory into memory.

        Returns:
            self.
        """
        data = self._get_trajectory_from_npz()
        self.data_keys = list(data.keys())
        for key in ["coords", "energies", "forces", "old_indices", "old_energies", "old_forces"]:
            self.assign_property(key, [x for x in data[key]])
        node_number = data["nuclear_charges"]
        self.assign_property("nuclear_charges", [np.array(node_number) for _ in range(len(self))])

        # Add splits to self.
        splits_train, splits_test = self._get_train_test_splits()
        property_train = []
        property_test = []
        for i in range(len(self)):
            is_train = []
            is_test = []
            for j, split in enumerate(splits_train):
                if i in split:
                    is_train.append(j + 1)
            for j, split in enumerate(splits_test):
                if i in split:
                    is_test.append(j + 1)
            # Add to list
            if len(is_train) > 0:
                property_train.append(np.array(is_train, dtype="int"))
            else:
                property_train.append(None)
            if len(is_test) > 0:
                property_test.append(np.array(is_test, dtype="int"))
            else:
                property_test.append(None)
        self.assign_property("train", property_train)
        self.assign_property("test", property_test)

        return self



import tensorflow as tf
from kgcnn.layers.conv.megnet_conv import MEGnetBlock
from kgcnn.layers.geom import NodeDistanceEuclidean, GaussBasisLayer, NodePosition, ShiftPeriodicLattice
from kgcnn.layers.modules import Dense, LazyAdd, Dropout, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingGlobalEdges, PoolingNodes
from kgcnn.layers.pool.set2set import PoolingSet2Set
from kgcnn.model.utils import update_model_kwargs

# from kgcnn.layers.casting import ChangeTensorType
ks = tf.keras

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2022.11.25"


# Implementation of Megnet in `tf.keras` from paper:
# Graph Networks as a Universal Machine Learning Framework for Molecules and Crystals
# by Chi Chen, Weike Ye, Yunxing Zuo, Chen Zheng, and Shyue Ping Ong*
# https://github.com/materialsvirtuallab/megnet
# https://pubs.acs.org/doi/10.1021/acs.chemmater.9b01294


model_default = {
    "name": "Megnet",
    "inputs": [{"shape": (None,), "name": "node_attributes", "dtype": "float32", "ragged": True},
               {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True},
               {"shape": [], "name": "graph_attributes", "dtype": "float32", "ragged": False}],
    "input_embedding": {"node": {"input_dim": 95, "output_dim": 64},
                        "graph": {"input_dim": 100, "output_dim": 64}},
    "make_distance": True, "expand_distance": True,
    "gauss_args": {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
    "meg_block_args": {"node_embed": [64, 32, 32], "edge_embed": [64, 32, 32],
                       "env_embed": [64, 32, 32], "activation": "kgcnn>softplus2"},
    "set2set_args": {"channels": 16, "T": 3, "pooling_method": "sum", "init_qstar": "0"},
    "node_ff_args": {"units": [64, 32], "activation": "kgcnn>softplus2"},
    "edge_ff_args": {"units": [64, 32], "activation": "kgcnn>softplus2"},
    "state_ff_args": {"units": [64, 32], "activation": "kgcnn>softplus2"},
    "nblocks": 3, "has_ff": True, "dropout": None, "use_set2set": True,
    "verbose": 10,
    "output_embedding": "graph",
    "output_mlp": {"use_bias": [True, True, True], "units": [32, 16, 1],
                   "activation": ["kgcnn>softplus2", "kgcnn>softplus2", "linear"]}
}


@update_model_kwargs(model_default)
def make_model(inputs: list = None,
               input_embedding: dict = None,
               expand_distance: bool = None,
               make_distance: bool = None,
               gauss_args: dict = None,
               meg_block_args: dict = None,
               set2set_args: dict = None,
               node_ff_args: dict = None,
               edge_ff_args: dict = None,
               state_ff_args: dict = None,
               use_set2set: bool = None,
               nblocks: int = None,
               has_ff: bool = None,
               dropout: float = None,
               name: str = None,
               verbose: int = None,
               output_embedding: str = None,
               output_mlp: dict = None
               ):
    r"""Make `MegNet <https://pubs.acs.org/doi/10.1021/acs.chemmater.9b01294>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.Megnet.model_default`.

    Inputs:
        list: `[node_attributes, edge_distance, edge_indices, state_attributes]`
        or `[node_attributes, node_coordinates, edge_indices, state_attributes]` if :obj:`make_distance=True` and
        :obj:`expand_distance=True` to compute edge distances from node coordinates within the model.

            - node_attributes (tf.RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_distance (tf.RaggedTensor): Edge attributes or distance of shape `(batch, None, D)` expanded
              in a basis of dimension `D` or `(batch, None, 1)` if using a :obj:`GaussBasisLayer` layer
              with model argument :obj:`expand_distance=True` and the numeric distance between nodes.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.
            - state_attributes (tf.Tensor): Environment or graph state attributes of shape `(batch, F)` or `(batch,)`
              using an embedding layer.
            - node_coordinates (tf.RaggedTensor): Node (atomic) coordinates of shape `(batch, None, 3)`.

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in :obj:`Embedding` layers.
        make_distance (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices. Expansion uses `gauss_args`.
        gauss_args (dict): Dictionary of layer arguments unpacked in :obj:`GaussBasisLayer` layer.
        meg_block_args (dict): Dictionary of layer arguments unpacked in :obj:`MEGnetBlock` layer.
        set2set_args (dict): Dictionary of layer arguments unpacked in `:obj:PoolingSet2Set` layer.
        node_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        edge_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        state_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        use_set2set (bool): Whether to use :obj:`PoolingSet2Set` layer.
        nblocks (int): Number of graph embedding blocks or depth of the network.
        has_ff (bool): Use feed-forward MLP in each block.
        dropout (int): Dropout to use. Default is None.
        name (str): Name of the model.
        verbose (int): Verbosity level of print.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    env_input = ks.Input(**inputs[3])

    # embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    uenv = OptionalInputEmbedding(**input_embedding['graph'],
                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
    edi = edge_index_input

    # Edge distance as Gauss-Basis
    if make_distance:
        x = xyz_input
        pos1, pos2 = NodePosition()([x, edi])
        ed = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ed = xyz_input

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Model
    vp = n
    ep = ed
    up = uenv
    vp = GraphMLP(**node_ff_args)(vp)
    ep = GraphMLP(**edge_ff_args)(ep)
    up = MLP(**state_ff_args)(up)
    vp2 = vp
    ep2 = ep
    up2 = up
    for i in range(0, nblocks):
        if has_ff and i > 0:
            vp2 = GraphMLP(**node_ff_args)(vp)
            ep2 = GraphMLP(**edge_ff_args)(ep)
            up2 = MLP(**state_ff_args)(up)

        # MEGnetBlock
        vp2, ep2, up2 = MEGnetBlock(**meg_block_args)(
            [vp2, ep2, edi, up2])

        # skip connection
        if dropout is not None:
            vp2 = Dropout(dropout, name='dropout_atom_%d' % i)(vp2)
            ep2 = Dropout(dropout, name='dropout_bond_%d' % i)(ep2)
            up2 = ks.layers.Dropout(dropout, name='dropout_state_%d' % i)(up2)

        vp = LazyAdd()([vp2, vp])
        ep = LazyAdd()([ep2, ep])
        up = ks.layers.Add()([up2, up])

    if use_set2set:
        vp = Dense(set2set_args["channels"], activation='linear')(vp)  # to match units
        ep = Dense(set2set_args["channels"], activation='linear')(ep)  # to match units
        vp = PoolingSet2Set(**set2set_args)(vp)
        ep = PoolingSet2Set(**set2set_args)(ep)
    else:
        vp = PoolingNodes()(vp)
        ep = PoolingGlobalEdges()(ep)

    ep = ks.layers.Flatten()(ep)
    vp = ks.layers.Flatten()(vp)
    final_vec = ks.layers.Concatenate(axis=-1)([vp, ep, up])

    if dropout is not None:
        final_vec = ks.layers.Dropout(dropout, name='dropout_final')(final_vec)

    # Only graph embedding for Megnet
    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `Megnet`.")

    main_output = MLP(**output_mlp)(final_vec)
    model = ks.models.Model(inputs=[node_input, xyz_input, edge_index_input, env_input], outputs=main_output)

    model.__kgcnn_model_version__ = __model_version__
    return model


model_crystal_default = {
    'name': "Megnet",
    'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
               {'shape': (None, 3), 'name': "node_coordinates", 'dtype': 'float32', 'ragged': True},
               {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True},
               {'shape': [1], 'name': "charge", 'dtype': 'float32', 'ragged': False},
               {'shape': (None, 3), 'name': "edge_image", 'dtype': 'int64', 'ragged': True},
               {'shape': (3, 3), 'name': "graph_lattice", 'dtype': 'float32', 'ragged': False}],
    'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                        "graph": {"input_dim": 100, "output_dim": 64}},
    "make_distance": True, "expand_distance": True,
    'gauss_args': {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
    'meg_block_args': {'node_embed': [64, 32, 32], 'edge_embed': [64, 32, 32],
                       'env_embed': [64, 32, 32], 'activation': 'kgcnn>softplus2'},
    'set2set_args': {'channels': 16, 'T': 3, "pooling_method": "sum", "init_qstar": "0"},
    'node_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
    'edge_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
    'state_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
    'nblocks': 3, 'has_ff': True, 'dropout': None, 'use_set2set': True,
    'verbose': 10,
    'output_embedding': 'graph',
    'output_mlp': {"use_bias": [True, True, True], "units": [32, 16, 1],
                   "activation": ['kgcnn>softplus2', 'kgcnn>softplus2', 'linear']}
}


@update_model_kwargs(model_crystal_default)
def make_crystal_model(inputs: list = None,
                       input_embedding: dict = None,
                       expand_distance: bool = None,
                       make_distance: bool = None,
                       gauss_args: dict = None,
                       meg_block_args: dict = None,
                       set2set_args: dict = None,
                       node_ff_args: dict = None,
                       edge_ff_args: dict = None,
                       state_ff_args: dict = None,
                       use_set2set: bool = None,
                       nblocks: int = None,
                       has_ff: bool = None,
                       dropout: float = None,
                       name: str = None,
                       verbose: int = None,
                       output_embedding: str = None,
                       output_mlp: dict = None
                       ):
    r"""Make `MegNet <https://pubs.acs.org/doi/10.1021/acs.chemmater.9b01294>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.Megnet.model_crystal_default`.

    Inputs:
        list: `[node_attributes, edge_distance, edge_indices, state_attributes, edge_image, lattice]`
        or `[node_attributes, node_coordinates, edge_indices, state_attributes, edge_image, lattice]`
        if :obj:`make_distance=True` and :obj:`expand_distance=True` to compute edge distances
        from node coordinates within the model.

            - node_attributes (tf.RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_distance (tf.RaggedTensor): Edge distance of shape `(batch, None, D)` expanded
              in a basis of dimension `D` or `(batch, None, 1)` if using a :obj:`GaussBasisLayer` layer
              with model argument :obj:`expand_distance=True` and the numeric distance between nodes.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.
            - node_coordinates (tf.RaggedTensor): Node (atomic) coordinates of shape `(batch, None, 3)`.
            - lattice (tf.Tensor): Lattice matrix of the periodic structure of shape `(batch, 3, 3)`.
            - edge_image (tf.RaggedTensor): Indices of the periodic image the sending node is located. The indices
                of and edge are :math:`(i, j)` with :math:`j` being the sending node.
            - state_attributes (tf.Tensor): Environment or graph state attributes of shape `(batch, F)` or `(batch,)`
                using an embedding layer.

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in :obj:`Embedding` layers.
        make_distance (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices. Expansion uses `gauss_args`.
        gauss_args (dict): Dictionary of layer arguments unpacked in :obj:`GaussBasisLayer` layer.
        meg_block_args (dict): Dictionary of layer arguments unpacked in :obj:`MEGnetBlock` layer.
        set2set_args (dict): Dictionary of layer arguments unpacked in `:obj:PoolingSet2Set` layer.
        node_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        edge_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        state_ff_args (dict): Dictionary of layer arguments unpacked in :obj:`MLP` feed-forward layer.
        use_set2set (bool): Whether to use :obj:`PoolingSet2Set` layer.
        nblocks (int): Number of graph embedding blocks or depth of the network.
        has_ff (bool): Use feed-forward MLP in each block.
        dropout (int): Dropout to use. Default is None.
        name (str): Name of the model.
        verbose (int): Verbosity level of print.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    env_input = ks.Input(**inputs[3])
    edge_image = ks.layers.Input(**inputs[4])
    lattice = ks.layers.Input(**inputs[5])

    # embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    uenv = OptionalInputEmbedding(**input_embedding['graph'],
                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
    edi = edge_index_input

    # Edge distance as Gauss-Basis
    if make_distance:
        x = xyz_input
        pos1, pos2 = NodePosition()([x, edi])
        pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice])
        ed = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ed = xyz_input

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Model
    vp = n
    ep = ed
    up = uenv
    vp = GraphMLP(**node_ff_args)(vp)
    ep = GraphMLP(**edge_ff_args)(ep)
    up = MLP(**state_ff_args)(up)
    vp2 = vp
    ep2 = ep
    up2 = up
    for i in range(0, nblocks):
        if has_ff and i > 0:
            vp2 = GraphMLP(**node_ff_args)(vp)
            ep2 = GraphMLP(**edge_ff_args)(ep)
            up2 = MLP(**state_ff_args)(up)

        # MEGnetBlock
        vp2, ep2, up2 = MEGnetBlock(**meg_block_args)(
            [vp2, ep2, edi, up2])

        # skip connection
        if dropout is not None:
            vp2 = Dropout(dropout, name='dropout_atom_%d' % i)(vp2)
            ep2 = Dropout(dropout, name='dropout_bond_%d' % i)(ep2)
            up2 = ks.layers.Dropout(dropout, name='dropout_state_%d' % i)(up2)

        vp = LazyAdd()([vp2, vp])
        ep = LazyAdd()([ep2, ep])
        up = ks.layers.Add()([up2, up])

    if use_set2set:
        vp = Dense(set2set_args["channels"], activation='linear')(vp)  # to match units
        ep = Dense(set2set_args["channels"], activation='linear')(ep)  # to match units
        vp = PoolingSet2Set(**set2set_args)(vp)
        ep = PoolingSet2Set(**set2set_args)(ep)
    else:
        vp = PoolingNodes()(vp)
        ep = PoolingGlobalEdges()(ep)

    ep = ks.layers.Flatten()(ep)
    vp = ks.layers.Flatten()(vp)
    final_vec = ks.layers.Concatenate(axis=-1)([vp, ep, up])

    if dropout is not None:
        final_vec = ks.layers.Dropout(dropout, name='dropout_final')(final_vec)

    # Only graph embedding for Megnet
    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `Megnet`.")

    main_output = MLP(**output_mlp)(final_vec)
    model = ks.models.Model(inputs=[node_input, xyz_input, edge_index_input, env_input, edge_image, lattice],
                            outputs=main_output)

    model.__kgcnn_model_version__ = __model_version__
    return model



import os
import numpy as np
from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset


class QM8Dataset(QMDataset, DownloadDataset):
    r"""Store and process QM8 dataset from `MoleculeNet <https://moleculenet.org/>`__ datasets.

    From `Quantum Machine <http://quantum-machine.org/datasets/>`__ :
    Due to its favorable computational efficiency, time-dependent (TD) density functional theory(DFT) enables
    the prediction of electronic spectra in a high-throughput manner across chemical space. Its predictions,
    however, can be quite inaccurate. We resolve this issue with machine learning models trained on deviations of
    reference second-order approximate coupled-cluster (CC2) singles and doubles spectra from TDDFT counterparts,
    or even from DFT gap. We applied this approach to low-lying singlet-singlet vertical electronic spectra of
    over 20000 synthetically feasible small organic molecules with up to eight CONF atoms. The prediction errors
    decay monotonously as a function of training set size. For a training set of 10000 molecules,
    CC2 excitation energies can be reproduced to within ±0.1 eV for the remaining molecules.
    Analysis of our spectral database via chromophore counting suggests that even higher accuracies can be achieved.
    Based on the evidence collected, we discuss open challenges associated with data-driven modeling of
    high-lying spectra and transition intensities.

    .. note::

        We take the pre-processed dataset from `MoleculeNet <https://moleculenet.org/>`_ .

    References:

        (1) L. Ruddigkeit, R. van Deursen, L. C. Blum, J.-L. Reymond, Enumeration of 166 billion organic small
            molecules in the chemical universe database GDB-17, J. Chem. Inf. Model. 52, 2864–2875, 2012.
        (2) R. Ramakrishnan, M. Hartmann, E. Tapavicza, O. A. von Lilienfeld, Electronic Spectra from TDDFT and
            Machine Learning in Chemical Space, J. Chem. Phys. 143 084111, 2015.

    """

    download_info = {
        "dataset_name": "QM8",
        "data_directory_name": "qm8",
        "download_url": "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb8.tar.gz",
        "download_file_name": 'gdb8.tar.gz',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "gdb8"
    }

    def __init__(self, reload: bool = False, verbose: int = 10):
        """Initialize QM8 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM8")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = [
            "E1-CC2", "E2-CC2", "f1-CC2", "f2-CC2", "E1-PBE0", "E2-PBE0", "f1-PBE0",
            "f2-PBE0", "E1-PBE0", "E2-PBE0", "f1-PBE0", "f2-PBE0", "E1-CAM", "E2-CAM", "f1-CAM", "f2-CAM"
        ]
        self.label_units = ["[?]"]*len(self.label_names)
        self.label_unit_conversion = np.array([1.0] * 14)  # Pick always same units for training
        self.dataset_name = "QM8"
        self.require_prepare_data = False
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name, self.unpack_directory_name)
        self.file_name = "qm8.csv"

        if not os.path.exists(self.file_path):
            original_name = os.path.join(self.data_directory, "qm8.sdf.csv")
            if os.path.exists(original_name):
                os.rename(original_name, self.file_path)

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)



from kgcnn.data.visual_graph import VisualGraphDataset


class VgdRbMotifsDataset(VisualGraphDataset):
    """
    Synthetic graph regression dataset containing 5000 small, randomly generated graphs,
    where some graphs were seeded with special red- and blue-dominant subgraph motifs. Blue values add
    a negative contribution to the graph's overall regression value, while red motifs contribute positively.
    These sub graph motifs present the ground truth explanations for this task.
    """

    def __init__(self):
        super(VgdRbMotifsDataset, self).__init__(name='rb_dual_motifs')



import numpy as np
from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class SIDERDataset(MoleculeNetDataset2018):
    r"""Store and process 'SIDER' dataset from `MoleculeNet <https://moleculenet.org/>`__ database.

    Compare reference:
    `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/moleculenet.html#freesolv-dataset>`__ reading:
    The Side Effect Resource (SIDER) is a database of marketed drugs and adverse drug reactions (ADR).
    The version of the SIDER dataset in DeepChem has grouped drug side effects into 27 system organ classes following
    MedDRA classifications measured for 1427 approved drugs.

    Random splitting is recommended for this dataset.

    The raw data csv file contains columns below:

        - “smiles”: SMILES representation of the molecular structure
        - “Hepatobiliary disorders” ~ “Injury, poisoning and procedural complications”: Recorded side effects for the
           drug. Please refer to `<http://sideeffects.embl.de/se/?page=98>`_ for details on ADRs.

    References:

        (1) Kuhn, Michael, et al. “The SIDER database of drugs and side effects.”
            Nucleic acids research 44.D1 (2015): D1075-D1079.
        (2) Altae-Tran, Han, et al. “Low data drug discovery with one-shot learning.”
            ACS central science 3.4 (2017): 283-293.
        (3) Medical Dictionary for Regulatory Activities. http://www.meddra.org/

    """

    def __init__(self, reload=False, verbose: int = 10):
        """Initialize SIDER dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        super(SIDERDataset, self).__init__("SIDER", reload=reload, verbose=verbose)

    def read_in_memory(self, **kwargs):
        super(SIDERDataset, self).read_in_memory(**kwargs)
        graph_labels = self.obtain_property("graph_labels")
        graph_labels = [np.array(x, dtype="float") for x in graph_labels]
        self.assign_property("graph_labels", graph_labels)



import tensorflow as tf
# import kgcnn.ops.activ
from kgcnn.layers.modules import Dense, Activation, Dropout
from kgcnn.layers.norm import GraphBatchNormalization, GraphLayerNormalization, GraphNormalization, \
    GraphInstanceNormalization
from kgcnn.layers.norm import global_normalization_args
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.relational import RelationalDense

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='MLPBase')
class MLPBase(GraphBaseLayer):
    r"""Base class for multilayer perceptron that consist of multiple feed-forward networks.

    This base class simply manages layer arguments for :obj:`MLP`. They contain arguments for :obj:`Dense`,
    :obj:`Dropout` and :obj:`BatchNormalization` or :obj:`LayerNormalization` or :obj:`GraphNormalization`
    since MLP is made up of stacked :obj:`Dense` layers with optional normalization and
    dropout to improve stability or regularization. Here, a list in place of arguments must be provided that applies
    to each layer. If not a list is given, then the single argument is used for each layer.
    The number of layers is determined by :obj:`units` argument, which should be list.

    Hence, this base class holds arguments for batch-normalization which should be applied between kernel
    and activation. And dropout after the kernel output and before normalization.

    This base class does not initialize any sub-layers or implements :obj:`call`, only for managing arguments.
    The actual :obj:`MLP` layer inherits from this class.

    """

    def __init__(self,
                 units,
                 use_bias=True,
                 activation=None,
                 activity_regularizer=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_initializer="glorot_uniform",
                 bias_initializer="zeros",
                 kernel_constraint=None,
                 bias_constraint=None,
                 # Normalization
                 use_normalization=False,
                 normalization_technique="BatchNormalization",
                 axis=-1,
                 momentum=0.99,
                 epsilon=0.001,
                 mean_shift=True,
                 center=True,
                 scale=True,
                 alpha_initializer="ones",
                 beta_initializer="zeros",
                 gamma_initializer="ones",
                 moving_mean_initializer="zeros",
                 moving_variance_initializer="ones",
                 alpha_regularizer=None,
                 beta_regularizer=None,
                 gamma_regularizer=None,
                 alpha_constraint=None,
                 beta_constraint=None,
                 gamma_constraint=None,
                 # Dropout
                 use_dropout=False,
                 rate=None,
                 noise_shape=None,
                 seed=None,
                 **kwargs):
        r"""Initialize with parameter for MLP layer that match :obj:`Dense` layer, including :obj:`Dropout` and
        :obj:`BatchNormalization` or :obj:`LayerNormalization` or :obj:`GraphNormalization` .

        Args:
            units: Positive integer, dimensionality of the output space.
            activation: Activation function to use.
                If you don't specify anything, no activation is applied
                (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean, whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix.
            bias_initializer: Initializer for the bias vector.
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix.
            bias_regularizer: Regularizer function applied to the bias vector.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation").
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix.
            bias_constraint: Constraint function applied to the bias vector.
            use_normalization: Whether to use a normalization layer in between.
            normalization_technique: Which keras normalization technique to apply.
                This can be either 'batch', 'layer', 'group' etc.
            axis: Integer, the axis that should be normalized (typically the features
                axis). For instance, after a `Conv2D` layer with
                `data_format="channels_first"`, set `axis=1` in `GraphBatchNormalization`.
            momentum: Momentum for the moving average.
            epsilon: Small float added to variance to avoid dividing by zero.
            mean_shift: Whether to apply alpha.
            center: If True, add offset of `beta` to normalized tensor. If False, `beta`
                is ignored.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used. When the
                next layer is linear (also e.g. `nn.relu`), this can be disabled since the
                scaling will be done by the next layer.
            alpha_initializer: Initializer for the alpha weight. Defaults to 'ones'.
            beta_initializer: Initializer for the beta weight.
            gamma_initializer: Initializer for the gamma weight.
            moving_mean_initializer: Initializer for the moving mean.
            moving_variance_initializer: Initializer for the moving variance.
            alpha_regularizer: Optional regularizer for the alpha weight.
            beta_regularizer: Optional regularizer for the beta weight.
            gamma_regularizer: Optional regularizer for the gamma weight.
            beta_constraint: Optional constraint for the beta weight.
            gamma_constraint: Optional constraint for the gamma weight.
            alpha_constraint: Optional constraint for the alpha weight.
            use_dropout: Whether to use dropout layers in between.
            rate: Float between 0 and 1. Fraction of the input units to drop.
            noise_shape: 1D integer tensor representing the shape of the
                binary dropout mask that will be multiplied with the input.
                For instance, if your inputs have shape`(batch_size, timesteps, features)` and
                you want the dropout mask to be the same for all timesteps,
                you can use `noise_shape=(batch_size, 1, features)`.
            seed: A Python integer to use as random seed.
        """
        super(MLPBase, self).__init__(**kwargs)
        local_kw = locals()
        # List for groups of arguments.
        self._key_list_act = [
            "activation", "activity_regularizer"
        ]
        self._key_list_dense = [
            "units", "use_bias", "kernel_regularizer", "bias_regularizer", "kernel_initializer", "bias_initializer",
            "kernel_constraint", "bias_constraint"
        ]
        self._key_list_norm_all = [
            "axis", "momentum", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer",
            "moving_mean_initializer", "moving_variance_initializer", "beta_regularizer",
            "gamma_regularizer", "beta_constraint", "gamma_constraint", "alpha_initializer", "alpha_regularizer",
            "alpha_constraint", "mean_shift"
        ]
        self._key_list_dropout = ["rate", "noise_shape", "seed"]
        self._key_list_use = ["use_dropout", "use_normalization", "normalization_technique"]
        self._key_list_init = [
            "kernel_initializer", "bias_initializer", "beta_initializer", "gamma_initializer",
            "moving_mean_initializer", "moving_variance_initializer", "alpha_initializer"
        ]
        self._key_list_reg = [
            "activity_regularizer", "kernel_regularizer", "bias_regularizer", "beta_regularizer", "gamma_regularizer",
            "alpha_regularizer"
        ]
        self._key_list_const = [
            "kernel_constraint", "bias_constraint", "beta_constraint", "gamma_constraint", "alpha_constraint"
        ]
        self._key_dict_norm = global_normalization_args

        # Summarize all arguments.
        self._key_list = []
        self._key_list += self._key_list_act + self._key_list_dense + self._key_list_norm_all + self._key_list_dropout
        self._key_list += self._key_list_use
        self._key_list = list(set(self._key_list))
        # Dictionary of kwargs for MLP.
        mlp_kwargs = {key: local_kw[key] for key in self._key_list}

        # Everything should be defined by units.
        if isinstance(units, int):
            mlp_kwargs["units"] = [units]
        if not isinstance(mlp_kwargs["units"], list):
            raise ValueError("Units must be a list or a single int for `MLP`.")
        self._depth = len(mlp_kwargs["units"])
        # Special case, if axis is supposed to be multiple axis, use tuple here.
        if not isinstance(axis, list):
            mlp_kwargs["axis"] = [axis for _ in range(self._depth)]
        # Special case, for shape, use tuple here.
        if not isinstance(noise_shape, list):
            mlp_kwargs["noise_shape"] = [noise_shape for _ in range(self._depth)]

        # Assert matching number of args
        def assert_args_is_list(args):
            if not isinstance(args, (list, tuple)):
                return [args for _ in range(self._depth)]
            return args

        # Make every argument to list.
        for key, value in mlp_kwargs.items():
            mlp_kwargs[key] = assert_args_is_list(value)

        # Check correct length for all arguments.
        for key, value in mlp_kwargs.items():
            if self._depth != len(value):
                raise ValueError(
                    "Provide matching list of units '%s' and '%s' or simply a single value." % (
                        mlp_kwargs["units"], key))

        # Deserialize initializer, regularizes, constraints and activation.
        for sl, sm in [
            (self._key_list_init, ks.initializers.get), (self._key_list_reg, ks.regularizers.get),
            (self._key_list_const, ks.constraints.get), (["activation"], ks.activations.get)
        ]:
            for key in sl:
                mlp_kwargs[key] = [sm(x) for x in mlp_kwargs[key]]

        # Fix synonyms for normalization layer.
        replace_norm_identifier = [
            ("batch", "BatchNormalization"), ("graph_batch", "GraphBatchNormalization"),
            ("layer", "LayerNormalization"), ("graph_layer", "GraphLayerNormalization"),
            ("graph", "GraphNormalization"), ("graph_instance", "GraphInstanceNormalization")
        ]
        for i, x in enumerate(mlp_kwargs["normalization_technique"]):
            for key_rep, key in replace_norm_identifier:
                if x == key_rep:
                    mlp_kwargs["normalization_technique"][i] = key

        # Assign separate '_conf_' for use keys.
        # All '_conf_' kwargs in '_conf_mlp_kwargs'.
        for key in self._key_list_use:
            setattr(self, "_conf_" + key, mlp_kwargs[key])
        self._conf_mlp_kwargs = mlp_kwargs

    def _get_conf_for_keys(self, key_list_to_fetch: list, name_postfix: str, i_layer: int):
        out_kwargs = {key: self._conf_mlp_kwargs[key][i_layer] for key in key_list_to_fetch}
        out_kwargs.update({"name": self.name + "_" + name_postfix + "_" + str(i_layer)})
        return out_kwargs

    def build(self, input_shape):
        """Build layer."""
        super(MLPBase, self).build(input_shape)

    def get_config(self):
        """Update config."""
        config = super(MLPBase, self).get_config()
        for key in self._key_list:
            config.update({key: self._conf_mlp_kwargs[key]})

        # Serialize initializer, regularizes, constraints and activation.
        for sl, sm in [
            (self._key_list_init, ks.initializers.serialize), (self._key_list_reg, ks.regularizers.serialize),
            (self._key_list_const, ks.constraints.serialize), (["activation"], ks.activations.serialize)
        ]:
            for key in sl:
                config.update({key: [sm(x) for x in self._conf_mlp_kwargs[key]]})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='MLP')
class MLP(MLPBase):
    r"""Multilayer perceptron that consist of multiple :obj:`Dense` layers.

    .. note::

        Please see layer arguments of :obj:`MLPBase` for configuration!

    This layer adds normalization and dropout for normal tensor input. Please, see keras
    `documentation <https://www.tensorflow.org/api_docs/python/tf>`_ of
    :obj:`Dense`, :obj:`Dropout`, :obj:`BatchNormalization` and :obj:`LayerNormalization` for more information.

    Additionally, graph oriented normalization is supported. You can choose :obj:`normalization_technique` to be
    either 'BatchNormalization', 'LayerNormalization', 'GraphLayerNormalization', or 'GraphBatchNormalization'
    or 'GraphNormalization' .

    """

    # If child classes want to replace layers.
    _supress_dense = False

    def __init__(self, units, **kwargs):
        """Initialize MLP. See MLPBase."""
        super(MLP, self).__init__(units=units, **kwargs)
        norm_classes = {
            "BatchNormalization": ks.layers.BatchNormalization,
            "GraphBatchNormalization": GraphBatchNormalization,
            "LayerNormalization": ks.layers.LayerNormalization,
            "GraphLayerNormalization": GraphLayerNormalization,
            "GraphNormalization": GraphNormalization,
            "GraphInstanceNormalization": GraphInstanceNormalization
        }
        if not self._supress_dense:
            self.mlp_dense_layer_list = [
                Dense(**self._get_conf_for_keys(
                    self._key_list_dense, "dense", i)) for i in range(self._depth)
            ]
        self.mlp_activation_layer_list = [
            Activation(**self._get_conf_for_keys(
                self._key_list_act, "act", i)) for i in range(self._depth)
        ]
        self.mlp_dropout_layer_list = [
            Dropout(**self._get_conf_for_keys(
                self._key_list_dropout, "drop", i)) if self._conf_use_dropout[i] else None for i
            in range(self._depth)
        ]
        self.mlp_norm_layer_list = [
            norm_classes[self._conf_normalization_technique[i]](
                **self._get_conf_for_keys(self._key_dict_norm[self._conf_normalization_technique[i]], "norm", i)
            ) if self._conf_use_normalization[i] else None for i in range(self._depth)
        ]

    def build(self, input_shape):
        """Build layer."""
        super(MLP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (tf.Tensor): Input tensor with last dimension not `None`.

        Returns:
            tf.Tensor: MLP forward pass.
        """
        x = inputs
        for i in range(self._depth):
            x = self.mlp_dense_layer_list[i](x, **kwargs)
            if self._conf_use_dropout[i]:
                x = self.mlp_dropout_layer_list[i](x, **kwargs)
            if self._conf_use_normalization[i]:
                x = self.mlp_norm_layer_list[i](x, **kwargs)
            x = self.mlp_activation_layer_list[i](x, **kwargs)
        out = x
        return out

    def get_config(self):
        """Update config."""
        config = super(MLP, self).get_config()
        return config


GraphMLP = MLP


@ks.utils.register_keras_serializable(package='kgcnn', name='RelationalMLP')
class RelationalMLP(MLP):
    r"""Multilayer perceptron that consist of multiple :obj:`Dense` layers.

    .. note::

        Please see layer arguments of :obj:`MLPBase` for configuration!

    This layer adds normalization and dropout for normal tensor input. Please, see keras
    `documentation <https://www.tensorflow.org/api_docs/python/tf>`_ of
    :obj:`Dropout`, :obj:`BatchNormalization` and :obj:`LayerNormalization` for more information.

    Additionally, graph oriented normalization is supported. You can choose :obj:`normalization_technique` to be
    either 'BatchNormalization', 'LayerNormalization', 'GraphLayerNormalization', or 'GraphBatchNormalization'
    or 'GraphNormalization' .

    """

    _supress_dense = True

    def __init__(self, units, num_relations: int, num_bases: int = None, num_blocks: int = None, **kwargs):
        """Initialize MLP. See MLPBase."""
        super(RelationalMLP, self).__init__(units=units, **kwargs)
        self._conf_num_relations = num_relations
        self._conf_num_bases = num_bases
        self._conf_num_blocks = num_blocks
        self._conf_relational_kwargs = {
            "num_relations": self._conf_num_relations, "num_bases": self._conf_num_bases,
            "num_blocks": self._conf_num_blocks
        }

        # Override dense list with RelationalDense layer.
        self.mlp_dense_layer_list = [RelationalDense(
            **self._conf_mlp_dense_layer_kwargs[i], **self._conf_relational_kwargs) for i in range(self._depth)]

    def build(self, input_shape):
        """Build layer."""
        super(RelationalMLP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [features, relation]

                - features (tf.Tensor, tf.RaggedTensor): Input tensor with last dimension not `None`.
                - relation (tf.Tensor, tf.RaggedTensor): Input tensor with last relation information.

        Returns:
            tf.Tensor: MLP forward pass.
        """
        x, relations = inputs
        for i in range(self._depth):
            x = self.mlp_dense_layer_list[i]([x, relations], **kwargs)
            if self._conf_use_dropout[i]:
                x = self.mlp_dropout_layer_list[i](x, **kwargs)
            if self._conf_use_normalization[i]:
                x = self.mlp_norm_layer_list[i](x, **kwargs)
            x = self.mlp_activation_layer_list[i](x, **kwargs)
        out = x
        return out

    def get_config(self):
        """Update config."""
        config = super(RelationalMLP, self).get_config()
        config.update(self._conf_relational_kwargs)
        return config



import typing as t

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.data.utils import ragged_tensor_from_nested_numpy


class AbstractExplanationMixin:

    def explain(self, x):
        raise NotImplementedError()


class ImportanceExplanationMixin:

    def explain(self, x, **kwargs):
        return self.explain_importances(x, **kwargs)

    # Returns a tuple of ragged tensors (node_importances, edge_importances)
    def explain_importances(self,
                            x: t.Sequence[tf.Tensor],
                            **kwargs
                            ) -> t.Tuple[tf.RaggedTensor, tf.RaggedTensor]:
        raise NotImplementedError


class AbstractExplanationMethod:

    def __call__(self, model, x, y):
        raise NotImplementedError


class ImportanceExplanationMethod(AbstractExplanationMethod):

    def __init__(self,
                 channels: int):
        self.channels = channels

    def __call__(self,
                model: ks.models.Model,
                x: tf.Tensor,
                y: tf.Tensor
                ) -> t.Tuple[tf.Tensor, tf.Tensor]:
        raise NotImplementedError


class MockImportanceExplanationMethod(ImportanceExplanationMethod):
    """
    This is a mock implementation of "ImportanceExplanationMethod". It is purely for testing purposes.
    Using this method will result in randomly generated importance values for nodes and edges.
    """
    def __init__(self, channels):
        super(MockImportanceExplanationMethod, self).__init__(channels=channels)

    def __call__(self,
                 model: ks.models.Model,
                 x: t.Tuple[tf.Tensor],
                 y: t.Tuple[tf.Tensor],
                 ) -> t.Tuple[tf.Tensor, tf.Tensor]:
        node_input, edge_input, _ = x

        # Im sure you could probably do this in tensorflow directly, but I am just going to go the numpy
        # route here because that's just easier.
        node_input = node_input.numpy()
        edge_input = edge_input.numpy()

        node_importances = [np.random.uniform(0, 1, size=(v.shape[0], self.channels))
                            for v in node_input]
        edge_importances = [np.random.uniform(0, 1, size=(v.shape[0], self.channels))
                            for v in edge_input]

        return (
            ragged_tensor_from_nested_numpy(node_importances),
            ragged_tensor_from_nested_numpy(edge_importances)
        )




import tensorflow as tf
import numpy as np
import math
from kgcnn.layers.geom import NodeDistanceEuclidean, NodePosition
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.modules import ExpandDims
from kgcnn.layers.gather import GatherNodesSelection
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.pooling import PoolingGlobalEdges, PoolingNodes

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='CENTCharge')
class CENTCharge(GraphBaseLayer):
    r"""Compute charge equilibration according to
    `Ko et al. (2021) <https://www.nature.com/articles/s41467-020-20427-2>`_ .

    The charge equilibration scheme seeks to minimize the energy :math:`{E}_{{\rm{Qeq}}}` of the following expression:

    .. math::

        {E}_{{\rm{Qeq}}}={E}_{{\rm{elec}}}+\mathop{\sum }\limits_{i=1}^{{N}_{{\rm{at}}}}({\chi }_{i}{Q}_{i}+
        \frac{1}{2}{J}_{i}{Q}_{i}^{2})\quad ,

    where :math:`{E}_{{\rm{elec}}}` is given as:

    .. math::

        {E}_{{\rm{elec}}}=\mathop{\sum }\limits_{i=1}^{{N}_{{\rm{at}}}}\mathop{\sum }\limits_{j<i}^{{N}_{{\rm{at}}}}
        \frac{{\rm{erf}}\left(\frac{{r}_{ij}}{\sqrt{2}{\gamma }_{ij}}\right)}{{r}_{ij}}{Q}_{i}{Q}_{j}+
        \mathop{\sum }\limits_{i=1}^{{N}_{{\rm{at}}}}\frac{{Q}_{i}^{2}}{2{\sigma }_{i}\sqrt{\pi }}
    
    with
    
    .. math::
    
        {\gamma }_{ij}=\sqrt{{\sigma }_{i}^{2}+{\sigma }_{j}^{2}}\quad .

    Where :math:`{J}_{i}` denotes the element-specific hardness and charge densities of width :math:`\sigma_i` taken
    from the covalent radii of the respective elements.
    To solve this minimization problem the derivatives with respect to the charges are set to zero:

    .. math::

        \frac{\partial {E}_{{\rm{Qeq}}}}{\partial {Q}_{i}}=0,\forall i=1,..,{N}_{{\rm{at}}}\ \Rightarrow \
        \sum\limits_{j=1}^{{N}_{{\rm{at}}}}{A}_{ij}{Q}_{j}+{\chi }_{i}=0

    with elements of :math:`{[{\bf{A}}]}_{ij}` defined like:

    .. math::

        {[{\bf{A}}]}_{ij}=\left\{\begin{array}{ll}{J}_{i}+\frac{1}{{\sigma }_{i}\sqrt{\pi }},&\,\text{if}\,\,
        i=j\\ \frac{{\rm{erf}}\left(\frac{{r}_{ij}}{\sqrt{2}{\gamma }_{ij}}\right)}{{r}_{ij}},&\,
        \text{otherwise}\,\end{array}\right.

    Finally, the linear equation system to solve with lagrange multiplier :math:`\lambda` reads:

    .. math::

        \begin{pmatrix} \begin{matrix}  \; & \; & \; \\ \; & {\bf{A}} & \; \\ \; & \; & \; \\ \end{matrix} &
        & \vert & \begin{matrix} 1 \\ \vdots \\ 1\end{matrix} \\ \hline
        \begin{matrix} 1 & \dots & 1\end{matrix} & & \vert & 0 \end{pmatrix} =
        \begin{pmatrix} Q_{1} \\ \vdots \\ Q_{N_{\rm{at}}} \\ \hline  \lambda \end{pmatrix} \cdot
        \begin{pmatrix} -\chi_1 \\ \vdots \\ \ -\chi_{N_{\rm{at}}} \\ \hline   Q_{\text{tot}} \end{pmatrix}

    A code example of using the layer and possible input is shown below:

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.conv.hdnnp_conv import CENTCharge
        layer = CENTCharge()
        z = tf.ragged.constant([[1, 6], [1, 1, 6]], ragged_rank=1)
        chi = tf.ragged.constant([[0.43, 0.37], [0.43, 0.43, 0.37]], ragged_rank=1)
        xyz = tf.ragged.constant([[[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
            [[0.0, 0.0, 0.0], [1.0, 0.2, 0.2], [2.0, 0.0, 0.0]]], ragged_rank=1, inner_shape=(3,))
        qtot = tf.constant([1.0, 1.0])
        q = layer([z, chi, xyz, qtot])
        print(q)

    """

    # From https://en.wikipedia.org/wiki/Covalent_radius with radii in pm.
    _default_radii = 0.01 * np.array([
        0.0, 31, 28,
        128, 96, 84, 73, 71, 66, 57, 58,
        166, 141, 121, 111, 107, 105, 102, 106,
        203, 176, 170, 160, 153, 139, 139, 132, 126, 124, 132, 122, 122, 120, 119, 120, 120, 116,
        220, 195, 190, 175, 164, 154, 147, 146, 142, 139, 145, 144, 142, 139, 139, 138, 139, 140,
        244, 215, 207, 204, 203, 201, 199, 198, 198, 196, 194, 192, 192, 189, 190, 187, 175, 187, 170, 162, 151, 144,
        141, 136, 136, 132, 145, 146, 148, 140, 150, 150,
        260, 221, 215, 206, 200, 196, 190, 187, 180, 169
    ])
    # Chemical hardness from https://www.pnas.org/doi/10.1073/pnas.2117416119 in eV
    _default_hardness = 0.037 / 0.529177 * np.array([
        0.0, 6.2, 8.8,
        2.2, 4.6, 3.8, 4.7, 7.1, 5.6, 6.1, 9.1,
        2.1, 4.0, 2.6, 3.3, 4.7, 3.8, 4.5, 7.7,
        2.3, 3.2, 3.2, 2.9, 3.2, 3.4, 4.0, 3.6, 3.3, 3.3, 3.8, 5.8, 3.0, 3.3, 4.5, 3.9, 4.2, 7.7,
        1.9, 3.1, 3.1, 2.9, 3.3, 3.5, 3.7, 3.7, 3.9, 4.1, 3.6, 5.4, 3.1, 3.1, 4.0, 3.6, 3.8, 6.8,
        1.8, 2.7, 2.4, 2.3, 2.5, 2.7, 2.5, 3.0, 3.0, 3.2, 3.2, 3.3, 3.3, 3.3, 3.1, 3.5, 3.2, 3.8, 3.1, 3.6, 3.7, 3.7,
        3.8, 3.5, 3.6, 5.8, 3.1, 3.4, 3.3, 3.6, 3.6, 6.1,
        1.8, 3.0, 2.8, 2.8, 3.1, 3.0, 3.1, 3.5, 3.3, 3.3
    ])
    _max_atomic_number = 97

    def __init__(self, output_to_tensor: bool = False, use_physical_params: bool = True,
                 param_constraint=None, param_regularizer=None, param_initializer="glorot_uniform",
                 param_trainable: bool = False, chi_negative: bool = False,
                 **kwargs):
        r"""Initialize :obj:`CENTCharge` layer.

        Args:
            output_to_tensor (bool): Whether to return padded tensor plus mask. Default is False.
            use_physical_params (bool): Whether to initialize weights with physical motivated parameters.
                Default is True.
            param_constraint: Constraints for parameters of this layer. Default is None.
            param_regularizer: Regularize for parameters of this layer. Default is None.
            param_initializer: Initializer for parameters of this layer. Default is None.
            param_trainable (bool): Whether to make layer parameters trainable. Default is False.
            chi_negative (bool): Whether chi input is negative or positive electro-negativities. Default is False.
        """
        super(CENTCharge, self).__init__(**kwargs)
        self.output_to_tensor = output_to_tensor
        self.use_physical_params = use_physical_params

        self.layer_cast_n = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="mask", boolean_mask=True)
        self.layer_cast_chi = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="mask")
        self.layer_cast_x = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="mask", boolean_mask=True)

        # We can do this in init since weights do not depend on input shape.
        self.param_initializer = ks.initializers.deserialize(param_initializer)
        self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
        self.param_constraint = ks.constraints.deserialize(param_constraint)
        self.param_trainable = param_trainable

        self.weight_j = self.add_weight(
            "hardness_j",
            shape=(self._max_atomic_number,),
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=self.param_trainable
        )
        self.weight_sigma = self.add_weight(
            "sigma",
            shape=(self._max_atomic_number,),
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=self.param_trainable
        )
        if self.use_physical_params:
            self.set_weights([self._default_hardness, self._default_radii])

        self.chi_negative = chi_negative

    def build(self, input_shape):
        super(CENTCharge, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass. Casts to padded tensor for :obj:`tf.linalg.solve()` .

        Args:
            inputs (list): [n, chi, xyz, qtot]

                - n (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - chi (tf.RaggedTensor): Learned electronegativities. Shape (batch, [N], 1)
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - qtot (tf.Tensor): Total charge per molecule of shape (batch, 1)

            mask (list): Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Charges of shape (batch, None, 1)
        """
        n, chi, x = self.assert_ragged_input_rank(inputs[:3], mask=mask, ragged_rank=1)
        qtot = inputs[3]
        if qtot.shape.rank > 1:
            qtot = tf.squeeze(qtot, axis=-1)

        num_atoms = n.row_lengths()

        # Cast to padded tensor for charge equilibrium.
        # Only keep mask for atomic number.
        n_pad, n_mask = self.layer_cast_n(n, **kwargs)
        chi_pad, _ = self.layer_cast_chi(chi, **kwargs)
        if not self.chi_negative:
            chi_pad = chi_pad * -1.0
        x_pad, _ = self.layer_cast_x(x, **kwargs)
        if chi_pad.shape.rank > 2:
            chi_pad = tf.squeeze(chi_pad, axis=-1)

        # Compute mask values for diagonal and off-diagonal.
        a_mask = tf.logical_and(tf.expand_dims(n_mask, axis=1), tf.expand_dims(n_mask, axis=2))
        a_diag_mask = tf.linalg.diag(n_mask)
        a_off_mask = tf.logical_and(a_mask, tf.logical_not(a_diag_mask))

        # Compute distance and gamma matrix elements.
        diff = tf.expand_dims(x_pad, axis=1) - tf.expand_dims(x_pad, axis=2)
        dist = tf.sqrt(tf.reduce_sum(tf.square(diff), axis=-1, keepdims=False))  # (batch, N, N)
        gamma_ij = tf.sqrt(tf.expand_dims(
            tf.square(tf.gather(self.weight_sigma, n_pad, axis=0)), axis=1) + tf.expand_dims(
            tf.square(tf.gather(self.weight_sigma, n_pad, axis=0)), axis=2))  # (batch, N, N)

        # Compute diagonal matrix elements.
        hardness_i = tf.gather(self.weight_j, n_pad, axis=0)  # (batch, N)
        sigma_i = tf.gather(self.weight_sigma, n_pad, axis=0)  # (batch, N)
        diag_part = tf.linalg.diag(hardness_i + 1.0 / sigma_i / tf.sqrt(math.pi))

        # Setup Matrix A.
        a = tf.where(
            a_off_mask,
            tf.math.divide_no_nan(
                tf.math.erf(tf.math.divide_no_nan(dist, gamma_ij * tf.sqrt(2.0))),
                dist
            ),
            tf.zeros_like(dist)
        )
        a = tf.where(
            a_diag_mask,
            diag_part,
            a
        )

        # Pad A and chi for Lagrange multipliers.
        a = tf.pad(a, [[0, 0], [0, 1], [0, 1]], mode='CONSTANT', constant_values=0)
        chi_pad = tf.pad(chi_pad, [[0, 0], [0, 1]], mode='CONSTANT', constant_values=0)

        # Add total charge and Lagrange multipliers to A.
        qtot_indices = tf.cast(tf.concat([
            tf.expand_dims(tf.range(tf.shape(chi_pad)[0], dtype=num_atoms.dtype), axis=-1),
            tf.expand_dims(num_atoms, axis=-1)
        ], axis=-1), dtype="int32")
        chi_pad = tf.tensor_scatter_nd_add(
            chi_pad,
            qtot_indices,
            qtot
        )
        idx = [tf.cast(tf.expand_dims(iter_tensor, axis=-1), dtype="int32") for iter_tensor in [
            tf.repeat(tf.range(tf.shape(num_atoms)[0]), num_atoms), tf.ragged.range(num_atoms).flat_values,
            tf.repeat(num_atoms, num_atoms)]]
        idx = tf.concat([tf.concat([idx[0], idx[1], idx[2]], axis=-1),
                         tf.concat([idx[0], idx[2], idx[1]], axis=-1)], axis=0)
        a = tf.tensor_scatter_nd_add(
            a,
            idx,
            tf.ones(tf.shape(idx)[0], dtype=a.dtype)
        )

        # Set diagonal for empty matrix
        idx_0 = tf.ragged.range(num_atoms + 1, tf.repeat(tf.shape(a)[2], tf.shape(num_atoms)[0]))
        idx_0 = tf.concat([tf.cast(tf.expand_dims(iter_tensor, axis=-1), dtype="int32") for iter_tensor in [
            idx_0.value_rowids(), idx_0.flat_values, idx_0.flat_values
        ]], axis=-1)
        a = tf.tensor_scatter_nd_add(
            a,
            idx_0,
            tf.ones(tf.shape(idx_0)[0], dtype=a.dtype)
        )

        # Check system to solve
        # return a, chi_pad

        # Compute charges with solve().
        charges = tf.linalg.solve(a, tf.expand_dims(chi_pad, axis=-1), adjoint=False, name=None)

        n_mask_pad = tf.pad(n_mask, [[0, 0], [0, 1]], mode='CONSTANT', constant_values=0)
        if self.output_to_tensor:
            # Padded tensor will have total charges in it.
            return charges, n_mask_pad

        # Make ragged tensor from padded charges.
        return tf.RaggedTensor.from_row_lengths(charges[n_mask_pad], num_atoms, validate=self.ragged_validate)

    def get_config(self):
        """Update layer config."""
        config = super(CENTCharge, self).get_config()
        config.update({
            "chi_negative": self.chi_negative,
            "output_to_tensor": self.output_to_tensor,
            "use_physical_params": self.use_physical_params,
            "param_constraint": ks.constraints.serialize(self.param_constraint),
            "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
            "param_initializer": ks.initializers.serialize(self.param_initializer),
            "param_trainable": self.param_trainable
        })
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='ElectrostaticEnergyCharge')
class ElectrostaticEnergyCharge(GraphBaseLayer):
    r"""Compute electric energy according to
    `Ko et al. (2021) <https://www.nature.com/articles/s41467-020-20427-2>`_ .

    where :math:`{E}_{{\rm{elec}}}` is given as:

    .. math::

        {E}_{{\rm{elec}}}=\frac{1}{2}\,
        \mathop{\sum }\limits_{i=1}^{{N}_{{\rm{at}}}}\mathop{\sum }\limits_{j\neq i}^{{N}_{{\rm{at}}}}
        \frac{{\rm{erf}}\left(\frac{{r}_{ij}}{\sqrt{2}{\gamma }_{ij}}\right)}{{r}_{ij}}{Q}_{i}{Q}_{j}+
        \mathop{\sum }\limits_{i=1}^{{N}_{{\rm{at}}}}\frac{{Q}_{i}^{2}}{2{\sigma }_{i}\sqrt{\pi }}
    
    with
    
    .. math::
    
        {\gamma }_{ij}=\sqrt{{\sigma }_{i}^{2}+{\sigma }_{j}^{2}}\quad .


    Where :math:`{J}_{i}` denotes the element-specific hardness and charge densities of width :math:`\sigma_i` taken
    from the covalent radii of the respective elements.

    However, here the indices of atoms for the pair contribution in the energy, are expected as graph-like indices.
    The factor :math:`\frac{1}{2}` can be controlled by multiplicity argument.
    Example of using this layer:

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.conv.hdnnp_conv import ElectrostaticEnergyCharge
        layer = ElectrostaticEnergyCharge()
        z = tf.ragged.constant([[1, 6], [1, 1, 6]], ragged_rank=1)
        q = tf.ragged.constant([[0.43, 0.37], [0.43, 0.43, 0.37]], ragged_rank=1)
        xyz = tf.ragged.constant([[[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
            [[0.0, 0.0, 0.0], [1.0, 0.2, 0.2], [2.0, 0.0, 0.0]]], ragged_rank=1, inner_shape=(3,))
        ij = tf.ragged.constant([[[0, 1], [1, 0]],
            [[0, 1], [0, 2], [1, 0], [1, 2], [2, 0], [2, 1]]], ragged_rank=1, inner_shape=(2,))
        eng = layer([z, q, xyz, ij])
        print(eng)

    """

    # From https://en.wikipedia.org/wiki/Covalent_radius with radii in pm.
    _default_radii = 0.01 * np.array([
        0.0, 31, 28,
        128, 96, 84, 73, 71, 66, 57, 58,
        166, 141, 121, 111, 107, 105, 102, 106,
        203, 176, 170, 160, 153, 139, 139, 132, 126, 124, 132, 122, 122, 120, 119, 120, 120, 116,
        220, 195, 190, 175, 164, 154, 147, 146, 142, 139, 145, 144, 142, 139, 139, 138, 139, 140,
        244, 215, 207, 204, 203, 201, 199, 198, 198, 196, 194, 192, 192, 189, 190, 187, 175, 187, 170, 162, 151, 144,
        141, 136, 136, 132, 145, 146, 148, 140, 150, 150,
        260, 221, 215, 206, 200, 196, 190, 187, 180, 169
    ])

    _max_atomic_number = 97

    def __init__(self, add_eps: bool = False, multiplicity: float = 2.0,
                 use_physical_params: bool = True, param_constraint=None, param_regularizer=None,
                 param_initializer="glorot_uniform", param_trainable: bool = False,
                 _suppress_weight_initialization: bool = False, **kwargs):
        r"""Initialize :obj:`ElectrostaticEnergyCharge` layer.

        Args:
            add_eps (bool): Whether to add epsilon. Default is False.
            multiplicity (float): Multiplicity in counting pair-energy terms. For undirected edges this will be 2,
                since for each edge is present two times. Default is 2.0 .
            use_physical_params (bool): Whether to initialize weights with physical motivated parameters.
                Default is True.
            param_constraint: Constraints for parameters of this layer. Default is None.
            param_regularizer: Regularize for parameters of this layer. Default is None.
            param_initializer: Initializer for parameters of this layer. Default is None.
            param_trainable (bool): Whether to make layer parameters trainable. Default is False.
            _suppress_weight_initialization (bool): Internal parameter to suppress weight initialization.
                Only used for child classes. Default is False.
        """
        super(ElectrostaticEnergyCharge, self).__init__(**kwargs)
        self.add_eps = add_eps
        self.multiplicity = multiplicity
        self._suppress_weight_initialization = _suppress_weight_initialization
        self.layer_pos = NodePosition(selection_index=[0, 1])
        self.layer_gather = GatherNodesSelection(selection_index=[0, 1])
        self.layer_dist = NodeDistanceEuclidean(add_eps=add_eps)
        self.layer_exp_dims = ExpandDims(axis=2)
        self.layer_pool_edges = PoolingGlobalEdges(pooling_method="sum")
        self.layer_pool_nodes = PoolingNodes(pooling_method="sum")

        # We can do this in init since weights do not depend on input shape.
        self._suppress_weight_initialization = _suppress_weight_initialization
        if not _suppress_weight_initialization:

            self.use_physical_params = use_physical_params
            self.param_initializer = ks.initializers.deserialize(param_initializer)
            self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
            self.param_constraint = ks.constraints.deserialize(param_constraint)
            self.param_trainable = param_trainable

            self.weight_sigma = self.add_weight(
                "sigma",
                shape=(self._max_atomic_number,),
                initializer=self.param_initializer,
                regularizer=self.param_regularizer,
                constraint=self.param_constraint,
                dtype=self.dtype, trainable=self.param_trainable
            )
            if self.use_physical_params:
                self.set_weights([self._default_radii])

    def build(self, input_shape):
        super(ElectrostaticEnergyCharge, self).build(input_shape)

    def _find_sigma_from_atom_number(self, inputs):
        return tf.gather(self.weight_sigma, inputs, axis=0)

    def _compute_gamma(self, inputs):
        sigma_i, sigma_j = inputs
        gamma_squared = tf.square(sigma_i) + tf.square(sigma_j)
        if self.add_eps:
            gamma_squared = gamma_squared + ks.backend.epsilon()
        return tf.sqrt(gamma_squared)

    @staticmethod
    def _compute_pair_energy(inputs):
        qi, qj, rij, gamma_ij = inputs
        frac = tf.math.divide_no_nan(tf.math.erf(tf.math.divide_no_nan(rij, gamma_ij * tf.sqrt(2.0))), rij)
        return qi*qj*frac

    @staticmethod
    def _compute_self_energy(inputs):
        q, sigma = inputs
        return tf.math.divide_no_nan(tf.square(q), sigma)/2.0/tf.sqrt(math.pi)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [n, q, xyz, ij]

                - n (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - q (tf.RaggedTensor): Learned atomic charges. Shape (batch, [N], 1)
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ij (tf.RaggedTensor): Edge indices of shape (batch, [N], 2)

            mask (list): Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.Tensor: Energy of shape (batch, 1)

        """
        n, q, xyz, ij = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
        if n.shape.rank <= 2:
            n = self.layer_exp_dims(n, **kwargs)
        if q.shape.rank <= 2:
            q = self.layer_exp_dims(q, **kwargs)
        xi, xj = self.layer_pos([xyz, ij], **kwargs)
        rij = self.layer_dist([xi, xj], **kwargs)
        ni, nj = self.layer_gather([n, ij], **kwargs)
        qi, qj = self.layer_gather([q, ij], **kwargs)
        sigma_i = self.map_values(self._find_sigma_from_atom_number, ni)
        sigma_j = self.map_values(self._find_sigma_from_atom_number, nj)
        sigma = self.map_values(self._find_sigma_from_atom_number, n)
        gamma_ij = self.map_values(self._compute_gamma, [sigma_i, sigma_j])
        pair_energy = self.map_values(self._compute_pair_energy, [qi, qj, rij, gamma_ij])
        self_energy = self.map_values(self._compute_self_energy, [q, sigma])
        sum_pair = self.layer_pool_edges(pair_energy)
        sum_self = self.layer_pool_nodes(self_energy)
        # Both are normal tensors now with shape (batch, 1).
        if self.multiplicity:
            sum_pair = sum_pair/self.multiplicity
        return sum_pair + sum_self

    def get_config(self):
        """Update layer config."""
        config = super(ElectrostaticEnergyCharge, self).get_config()
        config.update({
            "add_eps": self.add_eps,
            "multiplicity": self.multiplicity,
            "use_physical_params": self.use_physical_params,
            "param_constraint": ks.constraints.serialize(self.param_constraint),
            "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
            "param_initializer": ks.initializers.serialize(self.param_initializer),
            "param_trainable": self.param_trainable,
            "_suppress_weight_initialization": self._suppress_weight_initialization
        })
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='CENTChargePlusElectrostaticEnergy')
class CENTChargePlusElectrostaticEnergy(CENTCharge, ElectrostaticEnergyCharge):
    r"""Combines :obj:`CENTCharge` and :obj:`ElectrostaticEnergyCharge` .

    Please check documentation for :obj:`CENTCharge` and :obj:`ElectrostaticEnergyCharge` for information.

    """

    def __init__(self, output_to_tensor: bool = False, use_physical_params: bool = True,
                 param_constraint=None, param_regularizer=None, param_initializer="glorot_uniform",
                 param_trainable: bool = False, chi_negative: bool = False,
                 # For ElectrostaticEnergyCharge.
                 add_eps: bool = False, multiplicity: float = 2.0,
                 **kwargs):
        r"""Initialize :obj:`CENTChargePlusElectrostaticEnergy` layer.

        Args:
            output_to_tensor (bool): Whether to return padded tensor plus mask. Default is False.
            use_physical_params (bool): Whether to initialize weights with physical motivated parameters.
                Default is True.
            param_constraint: Constraints for parameters of this layer. Default is None.
            param_regularizer: Regularize for parameters of this layer. Default is None.
            param_initializer: Initializer for parameters of this layer. Default is None.
            param_trainable (bool): Whether to make layer parameters trainable. Default is False.
            chi_negative (bool): Whether chi input is negative or positive electro-negativities. Default is False.
            add_eps (bool): Whether to add epsilon. Default is False.
            multiplicity (float): Multiplicity in counting pair-energy terms. For undirected edges this will be 2,
                since for each edge is present two times. Default is 2.0 .
        """
        ElectrostaticEnergyCharge.__init__(
            self, add_eps=add_eps, multiplicity=multiplicity, _suppress_weight_initialization=True)
        CENTCharge.__init__(
            self, output_to_tensor=output_to_tensor, chi_negative=chi_negative, use_physical_params=use_physical_params,
            param_constraint=param_constraint, param_regularizer=param_regularizer,
            param_initializer=param_initializer, param_trainable=param_trainable, **kwargs)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass. Casts to padded tensor for :obj:`tf.linalg.solve()` .

        Args:
            inputs (list): [n, chi, xyz, ij, qtot]

                - n (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - chi (tf.RaggedTensor): Learned electronegativities. Shape (batch, [N], 1)
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ij (tf.RaggedTensor): Edge indices of shape (batch, [N], 2)
                - qtot (tf.Tensor): Total charge per molecule of shape (batch, 1)

            mask (list): Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tuple: Charges of shape (batch, [None], 1) plus energy of shape (batch, 1)
        """
        n, chi, xyz, ij, qtot = inputs
        q = CENTCharge.call(self, [n, chi, xyz, qtot], **kwargs)
        eng = ElectrostaticEnergyCharge.call(self, [n, q, xyz, ij], **kwargs)
        return q, eng

    def get_config(self):
        """Update layer config."""
        config = super(CENTChargePlusElectrostaticEnergy, self).get_config()
        config.update({
            "add_eps": self.add_eps,
            "multiplicity": self.multiplicity,
        })
        return config
